# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 30 11:43:48 2018

@author: huuskoj
"""

import unittest
from RHS import RHS
# 0  1   2   3    4     5    6   7    8     9    10
# A, It, Wt, Iy, Wely, Wply, iy, Iz, Welz, Wplz, iz
ARGS1 = [9.59*1e2, 215.8*1e4, 43.23*1e3, 195.8*1e4, 32.63*1e3,
         39.07*1e3, 45.2, 105.2*1e4, 26.30*1e3, 29.65*1e3, 33.1]
ARGS2 = [2.14*1e2,3.45*1e4, 2.36*1e3, 4.05*1e4, 2.02*1e3,
         2.61*1e3, 13.8, 1.34*1e4, 1.34*1e3, 1.60*1e3, 7.9]
# NcRd, McyRd, MczRd, VplyRd, VplzRd [kN /kNm]
RD1 = [299.5,13.87,8.02,78.61,117.9]
RD2 = [75.86,0.93,0.57,14.60,29.20]
"""
!!! NOTE !!!
rhs1 is cross-section class 4 and class 4 
calculations haven't been implemented yet
!!! NOTE !!!
"""
class TestRHS(unittest.TestCase):
    
    def test_area(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.A, ARGS1[0], delta=0.01*ARGS1[0])
        self.assertAlmostEqual(rhs2.A, ARGS2[0], delta=0.01*ARGS2[0])
    """
    def test_Wt(self):
        rhs1 = RHS(120,80,3)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.It, ARGS1[1], delta=1)
        self.assertAlmostEqual(rhs2.It, ARGS2[1], delta=1)
    """
    def test_It(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.It, ARGS1[2], delta=0.01*ARGS1[2])
        self.assertAlmostEqual(rhs2.It, ARGS2[2], delta=0.01*ARGS2[2])
    
    def test_Iy(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.I[0], ARGS1[3], delta=0.01*ARGS1[3])
        self.assertAlmostEqual(rhs2.I[0], ARGS2[3], delta=0.01*ARGS2[3])
    
    def test_Wely(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.Wel[0], ARGS1[4], delta=0.01*ARGS1[4])
        self.assertAlmostEqual(rhs2.Wel[0], ARGS2[4], delta=0.01*ARGS2[4])
        
    def test_Wply(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.Wpl[0], ARGS1[5], delta=0.01*ARGS1[5])
        self.assertAlmostEqual(rhs2.Wpl[0], ARGS2[5], delta=0.01*ARGS2[5])
    """   
    def test_iy(self):
        rhs1 = RHS(120,80,3)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.Wpl[1], ARGS1[6], delta=1)
        self.assertAlmostEqual(rhs2.Wpl[1], ARGS2[6], delta=1)
    """    
    def test_Iz(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.I[1], ARGS1[7], delta=0.01*ARGS1[7])
        self.assertAlmostEqual(rhs2.I[1], ARGS2[7], delta=0.01*ARGS2[7])
        
    def test_Welz(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.Wel[1], ARGS1[8], delta=0.01*ARGS1[8])
        self.assertAlmostEqual(rhs2.Wel[1], ARGS2[8], delta=0.01*ARGS2[8])
        
    def test_Wplz(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.Wpl[1], ARGS1[9], delta=0.01*ARGS1[9])
        self.assertAlmostEqual(rhs2.Wpl[1], ARGS2[9], delta=0.01*ARGS2[9])
    
    """
    def test_iz(self):
        rhs1 = RHS(120,80,3)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.Wpl[1], ARGS1[10], delta=1)
        self.assertAlmostEqual(rhs2.Wpl[1], ARGS2[10], delta=1)
    """   
       
    def test_NcRd(self):
        rhs1 = RHS(120,80,2.5)
        rhs1.Ned = -1
        rhs2 = RHS(40,20,2)
        rhs2.Ned = -1
        #self.assertAlmostEqual(rhs1.axial_force_resistance()/1e3, RD1[0], delta=1)
        self.assertAlmostEqual(rhs2.axial_force_resistance()/1e3, RD2[0], delta=1)
        
    def test_McyRd(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        #self.assertAlmostEqual(rhs1.bending_resistance()[0]/1e6, RD1[1], delta=1)
        self.assertAlmostEqual(rhs2.bending_resistance()[0]/1e6, RD2[1], delta=1)
        
    def test_MczRd(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        #self.assertAlmostEqual(rhs1.bending_resistance(axis='z')[0]/1e6, RD1[2], delta=1)
        self.assertAlmostEqual(rhs2.bending_resistance(axis='z')[0]/1e6, RD2[2], delta=1)
        
    def test_VcplyRd(self):
        rhs1 = RHS(120,80,2.5)
        rhs2 = RHS(40,20,2)
        self.assertAlmostEqual(rhs1.shear_force_resistance()/1e3, RD1[3], delta=1)
        self.assertAlmostEqual(rhs2.shear_force_resistance()/1e3, RD2[3], delta=1)       
        
        
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 14:59:36 2022

Constants to be used in optimization

@author: kmela
"""

CACHE_BOUND = 2**8
INT_TOL = 1e-4
""" Tolerance for discreteness violation of a discrete variable """
DISC_TOL = 1e-3
""" Tolerance for comparing variable vectors """
X_TOL = 1e-10

""" Artificial variable upper and lower bounds """
XUB = 1e8
XLB = -1e8

""" Tolerances for solvers """
GRAD_TOL = 1e-8  # Tolerance for the norm of gradient
ABS_F_TOL = 1e-8 # Tolerance for objective function difference, absolute
REL_F_TOL = 1e-4 # Tolerance for objective function difference, relative
G_TOL = 1e-8     # Tolerance for constraint feasibility
X_TOL = 1e-8     # Tolerance for change in variable values

POT_ACT_TOL = 0.25 # Tolerance for potentially active constraints


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Wed Oct 21 00:06:01 2020

CONLIN method

Convex-linear approximation of an optimization problem

@author: kmela
"""

import numpy as np
import time


from metku.optimization.solvers.optsolver import OptSolver
from metku.optimization.structopt import *
from metku.optimization.benchmarks import *

from ortools.linear_solver import pywraplp

class CONLIN(OptSolver):
    """ Solver class for Convex-Linear approximation Method (CONLIN) """
    
    def __init__(self, ):
        """ Constructor 
            :param move_limits: (tuple) relative lower and upper move limits
            :param gamma: parameter for reducing the relative move limits at each iteration
            :param C: penalty factor for safe-guarding againts potential infeasibility of LP problems
            :param beta: upper bound for the feasibility variable 'beta'. If beta = 0, then the feasibility variable
                        is ignored
            :param move_limit_type: type of move limits. 
                        Possible values: 
                            'range' .. move limits are taken with respect to the entire range of variable values. 
                            'relative' .. move limits are taken with respect to the current variable values.
        """
        super().__init__()
        self.move_limits = np.asarray(move_limits)
        self.move_limits_hist = [np.asarray(move_limits)]
        self.move_limit_type = move_limit_type
        self.gamma = gamma
        self.C = C
        self.beta = beta
        
        self.include_beta = True
        
        if beta == 0:
            self.include_beta = False

    def take_action(self):
        """
        Defines action to take
        Here, this means constructing the CONLIN approximation and solving it at x
        :return:
        """
        conlin_prob = self.problem.conlin(*self.X)
        # bounds = [x*self.move_limits for x in self.X]
        #
        # res = linprog(df, A, B, bounds=bounds)

        solver = pywraplp.Solver('SLP',
                                 pywraplp.Solver.CLP_LINEAR_PROGRAMMING)

        # Number of constraints
        n = A.shape[0]
        # Number of variables
        m = A.shape[1]

        x = {}
        # Variables
        for i, var in enumerate(self.problem.vars):
            # Delta is the range of variable values
            # First, lb and ub are the allowable changes in the variable value
            if self.move_limit_type == "range":
                delta = var.ub - var.lb        
                #lb, ub = self.move_limits * delta
            elif self.move_limit_type == 'relative':
                delta = var.value
                #lb = self.move_limits[0]*var.value
                #ub = self.move_limits[1]*var.value
            
            lb, ub = self.move_limits * delta
            """ Set variable bounds, taking into account true upper and lower bounds """
            lb = max(var.lb, var.value - lb)
            ub = min(var.ub, var.value + ub)
            x[i] = solver.NumVar(lb,
                                 ub,
                                 var.name)                     
        # Beta        
        if self.include_beta:
            beta = solver.NumVar(0, self.beta, 'beta')
        # Constraints
        for i in range(n):
            if self.include_beta:
                solver.Add(solver.Sum([A[i, j] * x[j] for j in range(m)]) <= B[i] + beta)
            else:
                solver.Add(solver.Sum([A[i, j] * x[j] for j in range(m)]) <= B[i])

        # Objective
        if self.include_beta:
            solver.Minimize(solver.Sum([df[j] * x[j] for j in range(m)]) + self.C * beta)
        else:
            solver.Minimize(solver.Sum([df[j] * x[j] for j in range(m)]))


        sol = solver.Solve()


        # If solution if infeasible
        if sol == 2:
            print("Solution found was infeasible!")
            # Return something != 0, otherwise iteration will
            # stop because two consecutive iterations produce
            # too similar results
            return np.zeros_like(self.X)

        if self.include_beta and beta.solution_value() < 1:
            self.beta = beta.solution_value()

        X = []
        for i in range(len(x)):
            X.append(x[i].solution_value())
        X = np.asarray(X)

        return X - self.X



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue May 21 19:24:45 2019

@author: kmela
"""

import math
from functools import partial
from metku.sections.steel.wi import WISection
#from structopt import Variable, LinearConstraint, Constraint, OptimizationProblem
import metku.structopt as sop
import numpy as np
import copy


def IBeamWeight(p,dvars,x):
    """ Weight of I-beam
        
        Parameters:
        -----------
            param:x: [h,tw,bf,tf]
            param:beam: SteelMember object
    
        Output:
        ------
            f: weight of the beam
            
    """
    
    #print('h = {0:4.2f}, tw = {1:4.2f}, bt = {2:4.2f}, bb = {3:4.2f}, tf1 = {4:4.2f}, tf2 = {5:4.2f}'.format(p.h,p.tw,p.b[0],p.b[1],p.tf[0],p.tf[1]))
    
    
    L = 6000
    for i in range(len(x)):
        dvars[i].substitute(x[i])
        
    #print('h = {0:4.2f}, tw = {1:4.2f}, bt = {2:4.2f}, bb = {3:4.2f}, tf1 = {4:4.2f}, tf2 = {5:4.2f}'.format(p.h,p.tw,p.b[0],p.b[1],p.tf[0],p.tf[1]))
    """
    p.h = x[0]
    p.tw = x[1]
    
    p.b[0] = x[2]
    p.b[1] = x[2]
    
    p.tf[0] = x[3]
    p.tf[1] = x[3]
    """
    
    return p.weight()*L

def IBeamWeightGrad(p,x):
    """ Gradient of the weight of I-beam
        
        Parameters:
        -----------
            param:x: [h,tw,bf,tf]
            param:beam: SteelMember object
    
        Output:
        ------
            f: weight of the beam
            
    """
    L = 6000
    """
    p.h = x[0]
    p.tw = x[1]
    
    p.b[0] = x[2]
    p.b[1] = x[2]
    
    p.tf[0] = x[3]
    p.tf[1] = x[3]
    """
    
    df = p.density*L*np.array([x[1],x[0]-2*x[3],2*x[3],2*x[2]-2*x[1]])
    
    return df

def IBeamWeightHessian(p,x):
    """ Hessian of the weight of I-beam
        
        Parameters:
        -----------
            param:x: [h,tw,bf,tf]
            param:beam: SteelMember object
    
        Output:
        ------
            f: weight of the beam
            
    """
    L = 6000
    """
    p.h = x[0]
    p.tw = x[1]
    
    p.b[0] = x[2]
    p.b[1] = x[2]
    
    p.tf[0] = x[3]
    p.tf[1] = x[3]
    """
    
    H = np.zeros([4,4])
    H[0,1] = 1
    H[1,0] = 1
    H[1,3] = -2
    H[2,3] = 2
    H[3,1] = -2
    H[3,2] = 2

    return p.density*L*H


def IBeamFlangeClassCon(p,flange_class=2):
    """ Builds a constraint for cross-section class of an I beam 
    
        0.5*b -  0.5*tw - sqrt(2)*aw <= C0*e*tf
    """
    
    a = [0,0,0,0]
    b = 0
    con_type = '<'
    
    e = p.eps
    
    if flange_class == 1:
        C0 = 9
    elif flange_class == 2:
        C0 = 10
    elif flange_class > 2:
        C0 = 14
    
    b = math.sqrt(2)*p.weld_throat
    a[1] = -0.5
    a[2] = 0.5
    a[3] = -C0*e
    
    
    #print(a)
    if flange_class > 3:
        """ if class 4 is required, the cf/tf ratio needs to
            be greater than the class 3 limit. This  changes the
            direction of the constraint from < to >.
        """
        con_type = '>'
    
    con_name = "Flange in class " + str(flange_class)
    con = sop.LinearConstraint(a,b,con_type,name=con_name)
    
    return con

def IBeamWebClassCon(p,web_class=2):
    """ Builds a constraint for cross-section class of an I beam 
    
        h - 2*tf -2*sqrt(2)*aw <= C1*e*tw
    """
    
    a = [0,0,0,0]
    b = 0
    con_type = '<'
    
    e = p.eps
     
    """ web is assumed to be in bending """
    if web_class == 1:
        C1 = 72
    elif web_class == 2:
        C1 = 83
    elif web_class > 2:
        C1 = 124
    
    b = 2*math.sqrt(2)*p.weld_throat
    a[0] = 1
    a[1] = -C1*e
    a[3] = -2
    
    #print(a)
    
    if web_class > 3:
        """ if class 4 is required, the cw/tw ratio needs to
            be greater than the class 3 limit. This  changes the
            direction of the constraint from < to >.
        """
        con_type[1] = '>'
    
    con_name = "Web in class " + str(web_class)
    con = sop.LinearConstraint(a,b,con_type,name=con_name)
    
    return con

def IBeamBendingCon(p,dvars,section_class,x):
    """ Constraint for the bending moment resistance """
    for i in range(len(x)):
        dvars[i].substitute(x[i])
    
    """
    p.h = x[0]
    p.tw = x[1]
    
    p.b[0] = x[2]
    p.b[1] = x[2]
    
    p.tf[0] = x[3]
    p.tf[1] = x[3]
    """
    
    if section_class < 3:
        MRd = p.plastic_bending_resistance()
    elif section_class == 3:
        MRd = p.elastic_bending_resistance()
    
    #print("Med = {0:4.2f}, MRd = {1:4.2f}".format(p.Med,MRd*1e-6))
    
    """ Med <= MRd -> Med/MRd >= 1 -> 1-Med/MRd <= 0 """
    
    return 1 - MRd*1e-6/p.Med
        

if __name__ == '__main__':
    
    p = WISection(h=300,tw=10,b=[100,100],tf=[15,15])
    p.Med = 20*6**2/8
    
    """ Design variables:
        x[0] .. height of the beam
        x[1] .. thickness of web
        x[2] .. width of the flanges
        x[3] .. thickness of flanges
    """
    dvars = []
    dvars.append(sop.Variable("Height",lb=150,ub=1000,target={'property':'H','objects':[p]}))
    dvars.append(sop.Variable("Web thickness",lb=5,ub=40,target={'property':'TW','objects':[p]}))
    dvars.append(sop.Variable("Flange width",lb=50,ub=500,target={'property':'BF','objects':[p]}))
    dvars.append(sop.Variable("Flange thickness",lb=5,ub=40,target={'property':'TF','objects':[p]}))
    #dvars.append(sop.Variable("Web thickness",5,40))
    
    obj = partial(IBeamWeight,p,dvars)
    obj_grad = partial(IBeamWeightGrad,p)
    obj_hess = partial(IBeamWeightHessian,p)
    
    prob = sop.OptimizationProblem(name="I-Beam Weight Minimization",\
                            variables=dvars,objective=obj,gradient=obj_grad,\
                            structure=p)
    
    flange_class = 2
    web_class = 2
    sec_class = max(flange_class,web_class)
    
    flange_con = IBeamFlangeClassCon(p,flange_class)
    web_con = IBeamWebClassCon(p,web_class)
    
    prob.add_constraint(flange_con)
    prob.add_constraint(web_con)
    
    bending_con_fun = partial(IBeamBendingCon,p,dvars,sec_class)
    bending_con = sop.NonLinearConstraint(bending_con_fun,'<',\
                                      "Bending resistance (mid-span)")
        

    prob.add_constraint(bending_con)
    
    x = [300,10,150,12]
    #prob(x)
    #print(prob.grad(x))
    #x0 = np.array(x)
    #print(sop.NumGrad(prob.obj,0.01*x0,x0))
    
    r0 = prob.solve("slsqp",x0=x)
    #r1 = prob.solve("trust-constr",x0=x)
    


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 19:11:28 2018

EN 1993-1-3 Supplementary rules for cold-formed members and sheeting

@author: kmela
"""

import math
import numpy as np

from metku.eurocodes.en1993.constants import gammaM0, gammaM1, gammaM2, gammaMfi
from metku.eurocodes.en1993 import en1993_1_2

def linear_interpolation(p1,p2,x):
    """ Performs linear interpolation.
        p1 = (x1,y1), p2 = (x2,y2)
        
        The function finds the value y corresponding to x1 <= x <= x2:
        Such that y(x) = k*x + b
    """
    
    y1 = p1[1]
    x1 = p1[0]
    y2 = p2[1]
    x2 = p2[0]
    
    k = (y2-y1)/(x2-x1)
    b = y1-k*x1
    
    return k*x+b

""" Chapter 3: Materials """
def fya(fyb,fu,t,Ag,n,k=7):
    """ Average yield strength, Eq. (3.1) 
        input:
            fyb .. basic yield strength [MPa]
            fu .. ultimate strength [MPa]
            k .. numerical coefficient depending on the type of forming
            n .. number of 90 degree bends
            t .. design core thickness of steel before colf-forming [mm]
    """

    return min(fyb+(fu-fyb)*k*n*t**2/Ag,0.5*(fu+fyb))

""" Chapter 5: Structural analysis """

def gr(r,t,phi):
    """ Calculate the length gr of Fig. 5.1 """
    rm = r + 0.5*t
    v = math.radians(phi)
    return rm*(math.tan(v) - math.sin(v))

def notional_width(b,rm,phi=np.array([90,90])):
    """ Notional width of a part
        input:
            b .. length of straight part of the part
            rm .. list or numpy array of midline radii of the corners
                 can be also a single number, if the part has only
                 one corner
            phi .. list or numpy array of corner angles.
                   can also be a single number
    """
    rm = np.array(rm)
    phi = np.array(phi)
    #print(sum(rm*np.sin(np.radians(0.5*phi))))
    return b + sum(rm*np.sin(np.radians(0.5*phi)))

def rounding_factor(r,phi,bp):
    """ Factor 'delta' of Eq. (5.1d) 
        input:
            r .. inner radius of different parts (list)
            phi .. angles (list)
            bp .. notional widths (list)
    """
    r = np.array(r)
    phi = np.array(phi)
    bp = np.array(bp)
    
    return 0.43*sum(r*phi/90)/sum(bp)

def distortional_buckling_reduction_factor(lambda_d):
    """ EN 1993-1-3, Eq. (5.12)
        Reduction factor for distortional buckling
        
        input:
            lambda_d .. slenderness = sqrt(fyb/sigma_cr,s)
    """
    
    if lambda_d <= 0.65:
        chi_d = 1.0
    elif lambda_d <= 1.38:
        chi_d = 1.47-0.723*lambda_d
    else:
        chi_d = 0.66/lambda_d
        
    return chi_d

def buckling_factor_edge_stiffener(bratio):
    """ EN 1993-1-3, Eq. (5.13b) and (5.13c)
        Buckling factor for single edge stiffener
        
        input:
            bratio = b_p,c/b_p, where
            b_p,c = notional width of the edge stiffener
            b_p = notional width of the flange etc.
    """
    
    if bratio <= 0.35:
        ksigma = 0.5
    elif bratio <= 0.6:
        ksigma = 0.5 + 0.83*((bratio-0.35)**2)**(1/3)
    else:
        print("Error: b ratio must be at most 0.6.")
        
    return ksigma

def distortional_buckling_stress(Is,As,E,K):
    """ EN 1993-1-3, Eq. (5.15)
        Critical stress for distortional buckling
    """    
    return 2*math.sqrt(K*E*Is)/As

""" Chapter 6: Design of members """
def axial_force_resistance(self,Ned,A,**kwargs):
    try:
        if Ned >= 0:
            NRd = kwargs['fya']*A/gammaM0        
        else:
            Aeff = kwargs['Aeff']        
            if Aeff < A:
                NRd = kwargs['fyb']*Aeff/gammaM0
            else:
                lambda_ratio = kwargs['lambda_ratio']
                fyb = kwargs['fyb']
                fya = kwargs['fya']
                NcRd = A*(fyb+4*(fya-fyb)*(1-lambda_ratio))
                    
                NRd = min(NcRd,fya*A)/gammaM0
    
    except:
        print("axial_force_resistance: invalid keyword!")
        NRd = 0.0
          
    return NRd

def RwRd(fyb,t,hw,ss,c,r,phi=90,e=0.0,flange_stiffeners=False,web_stiffeners=False,verb=False):
    """ Resistance to transverse forces, cross sections with a single
        web. EN 1993-1-3, 6.1.7.2
    

    Parameters
    ----------
    fyb : float
        Yield strength of the base material.
    t : float
        Design value of wall thickness.

    Returns
    -------
    RwRd : Resistance value, in N.

    """
    
    def k1(k):
        return 1.33-0.33*k
    
    def k2(r,t):
        return min(max(1.15-0.15*r/t,0.5),1.0)
    
    def k3(phi):
        return 0.7 + 0.3*(phi/90**2)

    def k4(k):
        return 1.22-0.22*k
    
    def k5(r,t):
        return min(1.06-0.06*r/t,1.0)
    
    def k5_star(k):
        return max(1.49-0.53*k,0.6)
    
    def k6(t):
        return 0.88-0.12*t/1.9
    
    def k7(hw,t):
        hw_t = hw/t
        
        if hw_t < 150:
            return 1+hw_t/750
        else:
            return 1.2
    
    def k8(hw,t,k):
        hw_t = hw/t
        
        if hw_t < 66.5:
            return 1/k
        else:
            return (1.1-hw_t/665)*k
    
    def k9(t):
        return 0.82 + 0.15*t/0.19
    
    def k10(hw,t,k):
        return (0.98-hw/t/865)*k
    
    def k11(t):
        return 0.64+0.31*t/1.9
    
    # Base value of the resistance
    Rw0 = t**2*fyb/gammaM1
    
    k = fyb/228
    
    if abs(e) < 1e-6:
        # In this case, the transverse load is acting at the end
        # of the member.
        if c <= 1.5*hw:
            if web_stiffeners:
                C0 = k7(hw,t)*(8.8+1.1*np.sqrt(ss/t))
            else:
                # Load is close to the end of the member
                C0 = k1(k)*k2(r,t)*k3(phi)
                ss_t = ss/t
                if flange_stiffeners:
                    C = C0*(9.04-hw/t/60)*(1+0.01*ss_t)
                else:
                    if ss_t <= 60:
                        C = C0*(5.92-hw/t/132)*(1+0.01*ss_t)
                    else:
                        C = C0*(5.92-hw/t/132)*(0.71+0.015*ss_t)
        else:
            if web_stiffeners:
                C0 = k5_star(k)*k6(t)*(13.2+2.87*np.sqrt(ss/t))                
            else:
                C0 = k3(phi)*k4(k)*k5(r,t)
                ss_t = ss/t
                if ss_t <= 60:
                    C = C0*(14.7-hw/t/49.5)*(1+0.007*ss_t)
                else:
                    C = C0*(14.7-hw/t/49.5)*(0.75+0.011*ss_t)
    else:
        # Two opposite forces.
        # Assume for now that the distance between forces is
        # e < 1.5*hw.
        # If this is not the case, then we go for case 1 and examine
        # the two forces separately.
        if c <= 1.5*hw:
            if web_stiffeners:
                C0 = k10(hw,t,k)*k11(t)*(8.8+1.1*np.sqrt(ss/t))
            else:
                # Load is close to the end of the member
                C0 = k1(k)*k2(r,t)*k3(phi)
                C = C0*(6.66-hw/t/64)*(1+0.01*ss/t)
        else:
            if web_stiffeners:
                C0 = k8(hw,t,k)*k9(t)*(13.2+2.87*np.sqrt(ss/t))
            else:
                C0 = k3(phi)*k4(k)*k5(r,t)
                C = C0*(21.0-hw/t/16.3)*(1+0.0013*ss/t)
            
    RwRd = C*Rw0
    
    return RwRd
    
def RwRd_two_webs(fyb,r,t,phi,hw,ss,verb=False,code='old',**kwargs):
    """    

    Parameters
    ----------
    fyb : float
        Yield strength of the base material [MPa].
    r : float
        internal corner radius.
    t : float
        thickness of the steel core
    phi : float
        angle of the web relative to flanges (degrees), 45 <= phi <= 90
    hw : float
        Height of the web.
    ss : float
        Support width.
    category : integer, optional
        Loading category. value is either 1 or 2. The default is 1.
    force_type : string
        Type of loading. Either 'load' for local load or 'support' for support
    **kwargs : 
        either empty, or 'e = ' or 'c = '.

    Returns
    -------
    Resistance [N]

    """
    # Young's modulus
    E = 210000
    
    e = -1
    c = -1
    
    for key, value in kwargs.items():        
        if key == 'e':
            e = value            
        elif key == 'c':
            c = value        
        elif key == 'VEd':
            # list of shear forces of transverse shear forces on each side
            # of the local load or support reaction, with
            # abs(VEd[0]) > abs(VEd[1])
            VEd = value
        else:
            raise ValueError('Incorrect input value, give either c or e.')
            
    category = 1
        
    # The last condition corresponds to internal support reaction, which means
    # that neither 'c' or 'e' is given as input parameter    
    if e > 1.5*hw or c > 1.5*hw or (e < 0 and c < 0):
        
        category = 2

        
    if category == 1:
        # This a value is for trapezoidal sheeting
        # for liner trays and hats, use a = 0.057
        if code == 'old':
            a = 0.075
        else:
            if e > 0:
                # Sheeting with opposite loading
                a = 0.075
            else:
                # Sheeting near clear end
                a = min(0.18*np.sqrt((c+ss)/1.5/hw),0.15)
        # Effective bearing length
        la = 10.0
    else:
        # This a value is for trapezoidal sheeting
        # for liner trays and hats, use a = 0.115
        a = 0.15
        
        if code == 'old':
            VEd = [abs(V) for V in VEd]
            
            beta_v = (VEd[0]-VEd[1])/(VEd[0]+VEd[1])
            
            if beta_v <= 0.2:
                la = ss
            elif beta_v >= 0.3:
                la = 10.0
            else:
                # linear intepolation
                la = ss - (beta_v-0.2)/0.1*(ss-10)
        else:
            la = ss
        
        la = min(la,200)
    
    RwRd = a*t**2*np.sqrt(fyb*E)*(1-0.1*np.sqrt(r/t))*(0.5+np.sqrt(0.02*la/t))*(2.4+(phi/90)**2)/gammaM1   
    
    if verb:
        print("Resistance for transverse loads:")
        print(f"Category: {category}")
        print(f"Web height: hw = {hw:{4}.{2}f} mm")
        print(f"Support width: ss = {ss:{4}.{2}f} mm")
        print(f"alpha = {a:{4}.{2}f}")
        
        if category == 2 and code == "old":
            print(f"beta_v = {beta_v:{4}.{2}f}")
        
        print(f"Effective bearing length: la = {la:{4}.{2}f} mm")
        
        print(f" Rw_Rd = {RwRd*1e-3:{4}.{2}f} kN")
        
        
    
    return RwRd

""" Chapter 8: Design of Joints """

""" Table 8.2: Design resistances for self-tapping screws """
def screw_validity(e1,e2,p1,p2,d,t,t1):
    """ Check the conditions on the range of validity """
    if e1 >= 3*d and e2 >= 1.5 and p1 >= 3*d and p2 >= 3*d and d <= 8.0 and d >= 3.0:
        return True
    else:
        return False

def screw_bearing_resistance(fu,d,t,t1,verb=False):
    """ Bearing resistance of a screw loaded in shear
        input:
            fu .. ultimate strength of the plate
            d .. diameter of the screw
            t .. thickness of the thinner plate
            t1 .. thickness of the thicker plate
            e1 .. the end distance from the centre of the fastener
                    to the adjacent end of the connected part, in 
                    the direction of load transfer
    """
    if t == t1:
        alfa = min(3.2*math.sqrt(t/d),2.1)
    elif t1 >= 2.5*t and t < 1.0:
        alfa = min(3.2*math.sqrt(t/d),2.1)
    elif t1 >= 2.5*t:
        alfa = 2.1
    else:
        # Linear interpolation not implemented!
        if verb:
            print("Linear interpolation for alpha")
        p0 = (t,min(3.2*math.sqrt(t/d),2.1))
        if t < 1.0:
            a1 = min(3.2*math.sqrt(t/d),2.1)
        else:
            a1 = 2.1
        p1 = (2.5*t,a1)
        
        print(p0,p1)
        alfa = linear_interpolation(p0,p1,t1)
        #alfa = 2.1
    # This is for rivets!
    #FbRdMax = fu*e1*t/1.2/gammaM2
        
    FbRd = alfa*fu*d*t/gammaM2 
    if verb:
        print("Self-tapping screw:")
        print("  t = {0:4.2f} mm".format(t))
        print("  t1 = {0:4.2f} mm".format(t1))
        print(" alpha = {0:4.2f}".format(alfa))
        print(" FbRd = {0:4.3f} kN".format(FbRd*1e-3))
    
    return FbRd, alfa

def screw_bearing_resistance_elevated_temp(fu,d,t,fub,e1,e2,T,verb=False):
    """ Bearing resistance at elevated temperatures """
    
    fuT = en1993_1_2.fyT(T,fu)
    kb = en1993_1_2.kbT(T)
    fubT = kb*fub

    ad = e1/3/d
    abT = min([ad,fubT/fuT,1.0])

    k1 = min(2*e2/d-1,2.5)
    
    FbRd = k1*abT*fubT*d*t/gammaMfi
    
    if verb:
        print("Self-tapping screw:")
        print(f"fuT = {fuT:.2f} MPa")
        print(f"kbT = {kb:.3f}")
        print(f"fubT = {fubT:.2f} MPa")
        print(f"d = {d} mm")
        print(f"t = {t} mm")
        print(f"alpha_(b,T) = {abT:.2f}")
        print(f"k1 = {k1:.2f}")
        print(" FbRd = {0:4.3f} kN".format(FbRd*1e-3))
    
    return FbRd

def screw_net_secton_resistance(Anet,fu):
    """ Net-section resistance
        input:
            Anet .. net cross-sectional area of the connected part
            fu ..  ultimate tensile strength of the supporting member 
                    into which a screw is fixed
    """
    FnRd = Anet*fu/gammaM2
    return FnRd
    
def screw_shear_resistance_fin(diameter,material="hardened"):
    """ Screw shear resistance according to the Finnish
        National Annex
        
        input:
            diameter .. diameter of the screw [mm]
            material .. either "hardened" or "stainless"
        
        output:
            shear strength of the screw [N]
    """
    
    """ Values from the table """
    FvRd_hardened = {4.8:5200, 5.5:7200, 6.3:9800, 8.0:16300}
    FvRd_stainless = {4.8:4600, 5.5:6500, 6.3:8500, 8.0:14300}
    
    if material == "hardened":        
        FvRd = FvRd_hardened[diameter]
    elif material == "stainless":
        FvRd = FvRd_stainless[diameter]
        
    return FvRd

def sheeting_shear_stiffness(t,s,hw,broof):
    """ Eq. (10.1b) 
        input:
            t .. design thickness of sheeting
            s .. distance between purlins
            hw .. height of sheeting (between centerlines)
            broof .. overall length of the roof diaphragm in the direction of the
                    span of the sheeting in mm.
    """
    return 1000*math.sqrt(t**3)*(50 + 10*broof**(1/3))*s/hw

def CD_A(ba,tnom,position,A,bT,bR,load='gravity'):
    """ EN 1993-1-3, Eq. (10.17) """
    if ba <= 125:
        kba = (ba/100)**2
    elif ba < 200:
        kba = 1.25*(ba/100)
    else:
        print("Warning: width of purlin flange must not exceed 200mm!")
        kba = 1.25*(ba/100)
        
    if tnom > 0.75:
        if position == 'pos':
            p = 1.1
        else:
            p = 1.5
    else:
        p = 1.5
    
    kt = (tnom/0.75)**p
        
    if bR <= 185:
        kbR = 1.0
    else:
        kbR = 185/bR
    
    if load == 'gravity':
        if position == 'pos':
            if tnom == 0.75:
                kA = 1.0+(A-1.0)*0.08
            elif tnom >= 1.0:
                kA = 1.0+(A-1.0)*0.095
        else:
            if tnom == 0.75:
                kA = 1.0+(A-1.0)*0.16
            elif tnom >= 1.0:
                kA = 1.0+(A-1.0)*0.095
    else:
        kA = 1.0
    
    """ Table 10.3
        Assume that pitch of fasteners is e = bR and sheet is fastened
        through troughs.
    """
    if load == 'gravity':    
        if position == 'pos':    
            bT_max = 40
            C100 = 5.2
        else:
            bT_max = 120
            C10 = 3.1
    else:
        C100 = 2.6
        bT_max = 40
        
    if bT > bT_max:
        kbT = math.sqrt(bT_max/bT)
    else:
        kbT = 1.0
    
    CDA = C100*kba*kt*kbR*kA*kbT
    
    return CDA

class GrooveStiffener():
    """ Class for a plane element with intermediate stiffener
        EN 1993-1-3 (2006): 5.5.3.3
    """
    
    def __init__(self,nodes,t):
        """ Contructor
            :param nodes: list of [y,z] coordinates of the points
            :param t: calculation thickness
        """
        self.nodes = nodes
        self.t = t
    
    def bp1(self):
        """ Determine the first notional width """

    def bp2(self):
        """ Determine the second notional width """

if __name__ == "__main__":
    
    from materials.steel_data import Steel    
    
    #plate = Steel("S350GD")
    
    #R = RwRd_two_webs(fyb=350, r=2.4, t=1.16, phi=65, hw=130, ss=150,verb=True,code='new',
    #                  c=1.6*130,VEd=[300,100])
    #print(R)
    
    d = 5.5
    t = 1.0-0.04
    FbRd = screw_bearing_resistance_elevated_temp(fu=420,d=d,t=t,fub=800,e1=109*d,e2=2.1*d,T=639,verb=True)    
    


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

from metku.optimization.structopt import *
from metku.frame2d.frame2d import *
from metku.frame2d.simple_truss import *


import numpy as np

FIFTEEN_BAR_AREAS_in2 = [0.111, 0.141, 0.174, 0.220, 0.270, 0.287, 0.347,
                         0.440, 0.539, 0.954, 1.081, 1.174, 1.333, 1.488,
                         1.764, 2.142, 2.697, 2.800, 3.131, 3.565, 3.813,
                         4.805, 5.952, 6.572, 7.192, 8.525, 9.300, 10.850,
                         13.300, 14.290, 17.170, 19.180]

FIFTEEN_BAR_AREAS_mm2 = [71.613, 90.968, 112.258, 141.935, 174.193, 185.161,
                         223.871, 283.87, 347.741, 615.483, 697.418, 757.418,
                         859.998, 959.998, 1138.062, 1381.933, 1739.997,
                         1806.448, 2019.996, 2299.995, 2459.995, 3099.994,
                         3839.992, 4239.992, 4639.991, 5499.989, 5999.988,
                         6999.986, 8580.628, 9219.336, 11077.397, 12374.169]
# Problem parameters
L = 9144  # mm
h = 3048 # mm
F = 444890  # N
E = 68950  # MPa
rho = 2768e-9  # kg/mm3

# Constraint limits
sigma_max = 172.37  # MPa # 25000 psi
delta_max = 50.8  # mm

class SimpleFifteenBarTruss(OptimizationProblem):

    # Problem parameters
    L = 9144    # mm
    F = 444890  # N
    E = 68950   # MPa
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 172.37  # MPa
    delta_max = 50.8    # mm

    def __init__(self, prob_type="discrete"):

        super().__init__(name='TenBarTruss')
        self.prob_type = prob_type
        self.structure = self.create_structure()
        self.create_variables(profiles=FIFTEEN_BAR_AREAS_mm2)
        self.create_constraints()
        self.create_objective()


    def create_variables(self, profiles=[0, 1e6]):

        """
        Creates variables used in optimization

        Appends each variable to 'vars' 'list from where they can be accessed
        """
        self.vars = []
        for i, mem in enumerate(self.structure.members.values()):
            name = 'A' + str(i + 1)
            if self.prob_type == "discrete":
                var = DiscreteVariable(name,
                                    values=profiles,
                                    target={"property": "A",
                                            "objects": [mem]})

            elif self.prob_type == "continuous":
                var = Variable(name,
                               lb=profiles[0],
                               ub=profiles[-1],
                               target={"property": "A",
                                       "objects": [mem]})

            elif self.prob_type == 'index':
                var = IndexVariable(name,
                                   values=profiles,
                                   target={"property": "A",
                                           "objects": [mem]})


            else:

                raise TypeError("Problem type must be either 'discrete' "
                                "or 'continuous")


            self.add(var)

    def create_structure(self):

        nodes = []
        nodes.append([0, h])
        nodes.append([L / 3, h])
        nodes.append([2 * L / 3, h])
        nodes.append([L, h])
        nodes.append([0, 0])
        nodes.append([L / 3, 0])
        nodes.append([2 * L / 3, 0])
        nodes.append([L, 0])

        mems = [[0, 1], [1, 2], [2, 3], [4, 5], [5, 6], [6, 7],
                [5, 1], [6, 2], [7, 3],
                [0, 5], [4, 1],
                [1, 6], [5, 2],
                [2, 7], [6, 3]]

        profs = []
        for i in range(len(mems)):
            profs.append('SHS 50x50x3.0')

        t = SimpleTruss(nodes, mems, profs)

        t.add(PointLoad(nodes[7], [0, -F, 0]))

        t.add(XYHingedSupport(nodes[0]))
        t.add(XYHingedSupport(nodes[4]))

        # Change material properties
        for mem in t.members.values():
            mem.material.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = FIFTEEN_BAR_AREAS_mm2[0]
            mem.rho = self.rho

        t.generate()
        t.calculate()
        return t

    def constraint_generator(self, mem):

        def compression_fun(x):
            return -mem.ned / (mem.A * mem.fy) - 1

        def tension_fun(x):
            return mem.ned / (mem.A * mem.fy) - 1

        def disp_fun(x):
            displacementsList = list(mem.nodal_displacements.values())
            displacementsMultipleArrays = displacementsList[0].values()

            max_vals = [max(array) for array in displacementsMultipleArrays]
            min_vals = [min(array) for array in displacementsMultipleArrays]

            max_val = max(max_vals)
            min_val = min(min_vals)

            abs_max = max(max_val, abs(min_val))
            return abs_max / self.delta_max - 1

        return compression_fun, tension_fun, disp_fun

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []
        i = 0
        for var in self.vars:
            for mem in var.target["objects"]:
                if isinstance(mem, FrameMember):
                    i += 1
                    compression_fun, tension_fun, disp_fun = self.constraint_generator(mem)

                    comp_con = NonLinearConstraint(con_fun=compression_fun,
                                                   name="Compression " + str(i),
                                                   )
                    comp_con.fea_required = True

                    tension_con = NonLinearConstraint(con_fun=tension_fun,
                                                      name="Tension " + str(i))
                    tension_con.fea_required = True

                    disp_con = NonLinearConstraint(con_fun=disp_fun,
                                                   name='Displacement ' + str(i),
                                                   )
                    disp_con.fea_required = True
                    self.add(comp_con)
                    self.add(tension_con)
                    self.add(disp_con)

    def create_objective(self):

        def objective(X):
            self.substitute_variables(X)
            weight = 0
            for mem in self.structure.members.values():
                weight += mem.weight
            return weight

        obj = ObjectiveFunction(name="Weight",
                                obj_fun=objective)

        self.add(obj)

def convert_to_Excel(list):
    '''
    Save cross section areas in result.xlsx - file.
    :param list: list[0] == weight of truss, list[1] == cross section areas
    '''
    # Importing pandas as pd
    import pandas as pd
    df = pd.DataFrame()

    # Creating column
    df['Cross sections of best truss'] = list[1]

    # Converting to excel
    df.to_excel('result.xlsx', index=False)

if __name__ == '__main__':

    from metku.optimization.solvers import ga
    from metku.optimization.solvers import pso
    from metku.optimization.solvers import aco

    problem = SimpleFifteenBarTruss('discrete')
    solver = aco.AntColonyOptimizer(pop_size=100)

    fopt, xopt = solver.solve(problem, maxiter=100)
    print(fopt, xopt)
    problem(xopt)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
import time

import numpy as np
import matplotlib.pyplot as plt

from metku.optimization.constants import GRAD_TOL, ABS_F_TOL, REL_F_TOL, G_TOL, X_TOL

class OptSolver:
    """
        Base class for optimization problem solvers
    """

    def __init__(self):
        self.constr_vals = np.array([-1])
        self.X = np.array([], dtype=float) # Current iterate
        self.problem = None   # OptimizationProblem class object, problem to be solved
        self.fvals = []       # List of objective function values at different iterations
        self.xvals = []       # List of iterates
        self.best_f = np.inf  # Best known objective function value
        self.best_x = None    # Best found design
        self.feasible = False # Flag to determine if a feasible solution has been found
        self.message = None
        self.constr_vals = None

    def calc_constraints(self, x=[]):
        """
        Calculates constraint values and saves them to numpy array

        Parameters:
        ------------
        :param x: (Optional, default: self.X) Point where constraint values are calculated

        """
        constr_vals = []
        if len(x) == 0:
            x = self.X
        # Calculate constraints
        for i in range(len(self.problem.cons)):
            constr_vals.append(self.problem.cons[i](x))
        self.constr_vals = np.asarray(constr_vals)
        return np.asarray(constr_vals)
    
    def is_feasible(self, x=[]):
        
        if len(x) == 0:
            constr_vals = self.constr_vals
        else:
            constr_vals = self.calc_constraints(x)
        
        res = True
        for g, con in zip(constr_vals,self.problem.cons):
            if con.type == '<':
                if g > self.problem.con_tol:
                    res = False
                    break
            elif con.type == '>':
                if g < -self.problem.con_tol:
                    res = False
                    break
            else:
                if abs(g) > self.problem.con_tol:
                    res = False
                    break
        return res

    def _create_eqcons(self):
        """ Creates a list of equality constraints

        :return: constraints
        """
        eqcons = []

        for con in self.problem.cons:
            if con.type == "=":
                eqcons.append(con)

        return eqcons

    def _create_ieqcons(self):
        """ Creates a list of inequality constraints

        :return: list of equality constraint functions
        """
        ieqcons = []

        for con in self.problem.cons:
            if con.type == "<" or con.type == "<=":
                ieqcons.append(con.neg_call)
            elif con.type == ">" or con.type == ">=":
                ieqcons.append(con)

        return ieqcons

    def _create_bounds(self):
        """ Creates a list of optimization variables' bounds

        :return: list of boundaries (lb, ub)
        """
        bounds = []
        for var in self.problem.vars:
            bound = (var.lb, var.ub)
            bounds.append(bound)

        return bounds

    def stopping_criterion(self,criterion='x',**kwargs):
        """ Check various stopping criteria.
            input:
                criterion .. string that states which criterion
                            is checked. Possible values:
                                'x' .. compare two design variable vectors
                                'f' .. compare two objective function values
        """

        res = False

        if criterion == 'x':
            xtol = X_TOL
            for key, value in kwargs.items():
                if key == 'x':
                    x = value
                elif key == 'xtol':
                    xtol = value

            if np.linalg.norm(x[1]-x[0]) <  xtol*np.linalg.norm(x[0]):
                res = True
                self.message = "Stopping criterion fulfilled: relative change to x smaller than tolerance."
        elif criterion == 'f':
            abs_f_tol = ABS_F_TOL
            rel_f_tol = REL_F_TOL

            for key, value in kwargs.items():

                if key == 'f':
                    f = value
                elif key == 'abs_f_tol':
                    abs_f_tol = value
                elif key == 'rel_f_tol':
                    rel_f_tol = value

            # f[1] .. objective function at x(k+1)
            # f[0] .. objective function at x(k)
            if abs(f[1]-f[0]) <= abs_f_tol + rel_f_tol*abs(f[0]):                
                res = True
                self.message = "Stopping criterion fulfilled: relative change to f smaller than tolerance."

        return res

    def take_action(self):
        """ This method needs to be implemented for each solver separately """
        pass

    def update_plot(self, fig, ax):
        """
        Updates the plotted figure
        
        NOTE: This method seems redundant, maybe remove it? (KMe, 6.10.2022)
        
        :return:
        """
        ax.clear()
        for mem in self.problem.structure.members.values():
            (x0, y0), (x1, y1) = mem.coordinates
            lw = mem.cross_section.A / 10000
            ax.plot((x0, x1), (y0, y1), 'k', linewidth=lw)
        #self.problem.structure.plot_loads()
        #self.problem.structure.plot_deflection(10, show=False)

        plt.title(f'Feasible: {self.problem.feasible()}')
        plt.axis('equal')
        fig.canvas.draw()
        plt.pause(0.0001)

    def step(self, action):
        """
        Takes a step
        :param action:
        :return:
        """

        self.step_length -= self.step_length * self.step_factor
        self.X += action
        for i in range(len(self.X)):
            self.X[i] = np.clip(self.X[i], self.problem.vars[i].lb,
                           self.problem.vars[i].ub)

        self.problem.substitute_variables(self.X)

        return self.X.copy(), 1, False, 'INFO'

    def solve(self, problem, x0=None, maxiter=-1, maxtime=-1, log=True,
              min_diff=1e-5, verb=False, plot=False):

        """
        Main iteration loop. Solves given problem

        :param problem:
        :param maxiter:
        :param maxtime:
        :return:
        """
        if maxiter < 0:
            maxiter = 100

        if maxtime < 0:
            maxtime = 1e9


        # Assign problem
        self.problem = problem
        # If initial starting point is/isn't defined        
        if x0 is not None:
            self.X = np.asarray(x0, dtype=float)
            problem.substitute_variables(x0)
        else:
            self.X = self.random_feasible_point()
            problem.substitute_variables(self.X)

        # Store initial point
        self.xvals.append(np.array(x0))
        self.fvals.append(problem.obj(x0))


        if plot:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            plt.ion()
            plt.show()

        # Assign done
        done = False
        # Start iteration
        t_total = 0
        for i in range(maxiter):
            #if verb:
            #    print("*** Iteration {0:g} ***".format(i+1))
            if plot:
                self.update_plot(fig, ax)
            # Check time
            start = time.time()
            t_0 = time.process_time()
            if t_total >= maxtime or done:
                print(t_total,maxtime,done)
                print("T_total - Stopping criterion fulfilled.")
                break
            # Save previous state
            prev_state = self.X.copy()
            # Define action to take. For each method, this is the
            # procedure for updating the design variables.
            # The take_action method needs to be implemented separately for
            # each method.
            action = self.take_action()
            # Take step
            state, reward, done, info = self.step(action)
                    
            # if (np.linalg.norm(prev_state - state)/np.linalg.norm(prev_state) <= min_diff):
            #    print("Sufficiently small change in variable values in consecutive iterations.")
            #    print(state,prev_state)
            #    done = True
            #     #break

            #if np.all(prev_state == state):
            #    print("No change in variable values.")
            #    done = True
                #break
            # Change current state
            self.X = state

            # Substitute new variables
            problem.substitute_variables(state)
            # Calculate constraints
            #print(self.X)
            #self.calc_constraints(self.X)
            state_feasible = self.is_feasible(self.X)
            print('Check state feasibility',state_feasible)
            #print(state)
            # Save best values

            fval = problem.obj(state)
            t_1 = time.process_time()
            t_total += t_1-t_0
            if verb:
                print(
                    f'\r Iteration {i + 1} / {maxiter}: Obj: {fval:.4f} Feasible: {state_feasible} '\
                    f'max g: {max(self.constr_vals):.4f}')
            if fval < self.best_f and state_feasible:
                self.best_f = fval
                self.best_x = state.copy()
                if verb:
                    #print(self.best_x)
                    if len(state) <= 10:
                        print(f"New best!: {fval:.3f} {[round(s, 3) for s in state]}")
                    else:
                        print(f"New best!: {fval:.3f}")
            # Log objective vals per iteration
            if log:
                #if verb:
                #    print("logging X = ",self.X)
                problem.num_iters += 1
                problem.fvals.append(problem.obj(self.X))
                problem.states.append(list(state))
                problem.gvals.append(list(self.constr_vals).copy())
                self.fvals.append(problem.obj(self.X))
                self.xvals.append(self.X)
            end = time.time()

            if verb:
                print(f"Iteration took: {end - start :.2f} s")

            # CHECK STOPPING CRITERIA
            # If new state is almost same as previous AND the design is feasible,
            # the iteration can be stopped.
            # TODO! Add Feasibility Check
            if self.is_feasible():
                if self.stopping_criterion('x', x=[prev_state, state], xtol=X_TOL):
                    print(self.message)
                    done = True
    
                if self.stopping_criterion('f', f=[problem.obj(prev_state),problem.obj(state)], 
                                           abs_f_tol=ABS_F_TOL, rel_f_tol=REL_F_TOL):
                    print(self.message)                
                    done = True

            if done:
                break

        if self.best_x is None:
            self.best_x = self.X
            self.best_f = problem.obj(self.X)

        return self.best_f, self.best_x    

    def random_feasible_point(self):
        """
        Creates a random feasible starting point for optimization

        :return: random feasible starting point
        """
        print("Creating random feasible starting point!")

        X = [0] * self.problem.nvars()
        for i, var in enumerate(self.problem.vars):
            if self.problem.prob_type == 'continuous':
                X[i] = var.lb + (var.ub - var.lb) * np.random.rand()
            else:
                X[i] = np.random.randint(var.lb, var.ub)
        while np.any(self.calc_constraints(X) > -0.15):

            for i, var in enumerate(self.problem.vars):
                if self.problem.prob_type == 'continuous':
                    X[i] = var.lb + (var.ub - var.lb) * np.random.rand()
                else:
                    X[i] = np.random.randint(var.lb, var.ub)

            self.problem.substitute_variables(X)
            self.problem.fea()

        print("Starting point created!")

        return np.asarray(X)

    def plot_iters(self):
        """
        Plots iteration history

        Returns
        -------
        None.

        """
        
        niter = len(self.fvals)
        
        if niter > 0:        
            fig, ax = plt.subplots(1)
            it = np.arange(niter)
            ax.plot(it,self.fvals)
            
            ax.set_xlabel('Iteration')
            ax.set_ylabel('Objective function')
            
            ax.grid(True,axis='both',linestyle='--')
        
        


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
import csv
import os

from datetime import datetime

def to_csv(problem, solver, name=""):
    """
    Saves problem's results as a csv file

    :param problem: optimized problem
    :type problem: OptimizationProblem
    :return:
    """
    if not name:
        time = datetime.now().strftime('_%Y_%m_%d')
        name = type(solver).__name__ + '_' + problem.name + time + '.csv'

    # Optimal variable values
    xopt = [round(var.value,2) for var in problem.vars]

    # Optimal result
    fopt = problem.obj(xopt)

    # Constraints' values
    con_vals = problem.eval_cons(xopt)

    # Initial point, x0
    x0 = [round(var.value,2) for var in problem.x0]

    # No. Iterations
    iters = problem.num_iters

    # No. FEM evaluations
    fem_analyses = problem.num_fem_analyses

    # Avg. subproblem iterations / time

    # States
    states = problem.states

    # Function values
    fvals = problem.fvals


    results = {'f*': fopt,
               'x*': xopt,
               'x0': x0,
               'Iterations': iters,
               'FEM Analyses': fem_analyses
               }

    write_header = not os.path.exists(name)

    with open(name, 'a', newline='') as csvfile:
        fieldnames = list(results.keys())

        writer = csv.DictWriter(csvfile, fieldnames)
        if write_header:
            writer.writeheader()
        writer.writerow(results)


    #print(name, fopt, xopt, x0, iters, fem_analyses)

    print(f'{name} file created to {os.getcwd()}')


if __name__ == '__main__':
    from metku.optimization.benchmarks import *
    from metku.optimization.solvers import *
    solver = SLP(move_limits=[0.5, 1.5])
    maxiters = [10, 20, 30, 40, 50]
    for maxiter in maxiters:
        problem = ThreeBarTruss('continuous')
        solver.solve(problem, maxiter=maxiter, log=True)
        to_csv(problem, solver)





# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Sun Apr 10 16:21:20 2022

Plane truss example using raami package

@author: Kristo Mela
"""

from metku.raami.raami_plane_truss import Ktruss_example, Ntruss_example
from metku.raami.raami_plane_truss import SlopedTruss
from metku.raami.exports import AbaqusOptions
from metku.raami.raami_truss_opt import minimize_eccentricity
from metku.optimization.solvers.slsqp import SLSQP
from metku.optimization.solvers.trust_region import TrustRegionConstr

def LauriKTruss(span,h2,h1,dx,nel_chord=4,nel_brace=4,ndiv=4):
    # Create K truss for Lauri
    t = SlopedTruss(L1=0.5*span,L2=0.5*span,h2=h2,h1=h1,dx1=dx,dx2=dx)
    t.braces_as_beams = True
    t.generate_topology('K',ndiv,nel_chord=nel_chord,nel_brace=nel_brace)
    t.generate_supports()    
    t.generate_joints()    
    t.symmetry()
    t.generate_uniform_load(q=-25)
    t.generate_fem(model='no_eccentricity')
        
    print("Structural analysis..")
    t.structural_analysis(load_id=t.load_ids[0],support_method="REM")
    print("Done.")
    
    top = {'material': 'S355', 'class': 2, 'utility': 1.0}
    bottom = {'material': 'S355', 'class': 2, 'utility': 1.0}
    braces = {'material': 'S355', 'class': 2, 'utility_tens': 1.0, 'utility_comp':1.0}
    
    t.optimize_members(verb=True,top=top,bottom=bottom,braces=braces,limit_width=True)
     
    P, x0 = minimize_eccentricity(t,min_gap=20)   
    
        
    solver = SLSQP()
    #solver = TrustRegionConstr()
    #min_ecc, xmin = solver.solve(P,x0=x0,verb=True)
    #t.plot(geometry=True,loads=False)
    
    #t.clear_fem()
    #t.generate_fem(model="ecc_elements")
    #t.fem.draw()
    
    #t.print_member_utilization('Utilization.txt')
    
    #fig_save_opts = {'filename':'default.pdf','format':'pdf','orientation':'landscape','papertype':'a3'}    
    #t.plot(geometry=True,loads=False,save=True,saveopts=fig_save_opts)
    
    #t.generate_fem(model='ecc_elements')
    #opts = AbaqusOptions(x_monitor = 0.5*t.span, n_monitored = 2)
    #t.to_abaqus(filename='K-ristikko',partname="K-ristikko",options=opts)
    
    
    return t
    

if __name__ == "__main__":
    
    t = LauriKTruss(span=24000,h2=2400,h1=1800,dx=1000,nel_chord=6,nel_brace=6,ndiv=4)
    #t =  Ktruss_example(h2=2000,h1=1500,dx1=1000,dx2=1000,first=False,edges=True)
    #t =  Ntruss_example(h2=2000,h1=1500,dx1=1500,dx2=1500,first=False,edges=True)
    
    #t.generate_fem(model='en1993')
    #t.generate_fem(model='no_eccentricity')
    #t.structural_analysis(load_id=t.load_ids[0],support_method="REM")
    #x, xc = t.joints[4].brace_chord_face_x()
    #t.bmd(scale=10,load_id=t.load_ids[0])

#t.plot(geometry=True,loads=False)



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 15:21:39 2022

Objective function class for optimization problems

@author: kmela
"""

import numpy as np
from metku.optimization.structopt import OptimizationProblem
from optimzation_enums import ObjectiveTypeEnum


class ObjectiveFunction:

    def __init__(self,
                 name: str,
                 obj_fun: callable,
                 obj_type: ObjectiveTypeEnum = ObjectiveTypeEnum.MIN,
                 problem: OptimizationProblem = None,
                 fea_required: bool = False):

        self.name = name
        if not callable(obj_fun):
            raise ValueError("obj_fun must be a function!")
        self.obj_fun = obj_fun
        self.obj_type = obj_type[:3].upper()
        self.problem = problem
        self.fea_required = fea_required

    def __call__(self, x):
        """ The method returns f(x) where x is an array with length
            of the number of free variables. The array will be appended
            for any fixed variables with corresponding fixed values.
            This way, if some variables are fixed, the original objective
            function can still be called.

            Read values from 'x' and reverse them.
            Reversing is done so that 'pop' method can be employed.
        """
        if self.fea_required and not self.problem.fea_done:
            self.problem.fea()

        X = [val for val in x]
        X.reverse()

        """ Fixed values values is an array
            with the values
            fixed_vals[i] is None if the variable vars[i] is free and
            fixed_vals[i] is val if the variable vars[i] has a fixed value val
        """
        fixed_vals = self.problem.fixed_vals.copy()

        for i, val in enumerate(fixed_vals):
            if val is None:
                """ If variable vars[i] is free, take its value from X
                    Because X was reversed, the pop method provides the correct
                    value here.
                """
                fixed_vals[i] = X.pop()
            if not X:
                """ If variable vars[i] is fixed, then use the corresponding
                    value from 'fixed_vals'
                """
                break

        self.problem.substitute_variables(fixed_vals)

        if self.obj_type == ObjectiveTypeEnum.MIN:
            return self.obj_fun(fixed_vals)
        else:
            return -self.obj_fun(fixed_vals)

    def neg_call(self, x):
        """
        Calculates negative value for objective function

        :return: negative value of objective function
        """
        return -self(x)


class LinearObjective(ObjectiveFunction):
    """ Class for linear objective functions """

    def __init__(self, name, c, obj_type="MIN", problem=None):
        """ Constructor:
            input:
                c .. constant array for performing the evaluation
                     c @ x
        """
        obj_fun = self.evaluate

        self.c = np.array(c)

        super().__init__(name, obj_fun, obj_type, problem)

    def evaluate(self, x):
        """ Evaluate function value """
        return self.c.dot(np.array(x))



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
""" Particle swarm optimization
"""

import operator
import random
import copy

import matplotlib.pyplot as plt
import numpy as np
from deap import base, creator, tools


from metku.optimization.solvers.optsolver import OptSolver, G_TOL


class PSO(OptSolver):
    """ Particle swarm optimization solver """
    def __init__(self,pop_size=10, c = [2.0,2.0], inertia=0.9, crazy_prob=0.1, 
                 penalty=None,best_f = np.inf):
        """ Constructor
            :param pop_size: population size
            :c: [c1,c2] weighting factors
            :inertia: initial value for the inertia term
            :craziness: craziness probability            
        
        """
        
        super().__init__()

        self.pop_size = pop_size
        self.c = c
        self.inertia = inertia
        self._craziness = crazy_prob
        self.plot = False
        self.best_x = None
        self.best_f = best_f
        self.best_scv = np.inf
        self.prev_best = best_f
        self.best_part = None
        self.swarm = [0 for _ in range(pop_size)]
        self.vrel = 0.2
        self.vmin = None
        self.vmax = None
        self.vred = 0.95
        self.NMAX = 10
        self.KMAX = 5
        if penalty is None:
            def penalty(constr_vals):
                pos_vals = np.clip(constr_vals, 0, 100)
                return 2e6 * sum(pos_vals ** 2)

            self.penalty = penalty
        else:
            self.penalty = penalty

        self.counter = 0
        self.inertia_cnt = 0

    def eval_particle(self, part):
        """
        Evaluates one particle

        :param part: Particle class object
        """
        self.problem.substitute_variables(part.x)
        #X = [round(var.value, 5) for var in self.problem.vars]
        
        """ Evaluate objective function """
        part.fval = self.problem.obj(part.x)
        """ Evaluate penalty function """
        part.scv = self.penalty(part.x)
        
                    
        """ If the particle is currently infeasible, it is sufficient to
            check if the 'scv' value is smaller than the current value.
            
            NOTE! This part needs to be changed if penalty function approach
            is used.
        """
        if part.best_scv > G_TOL:
            #print("Particle is currently infeasible")
            if part.scv <= part.best_scv:
                #print("Improving...")
                part.update_best()
        else:
            """ If the particle is currently feasible, we need to check
                both the scv and val values
            """
            if part.scv <= G_TOL and part.fval <= part.best_f:
                part.update_best()
                
        if self.plot:
            self.update_plot(self.fig, self.ax)
        

    def update_best(self):
        """ Update current best solution 
            Assumption is that all particle of the swarm have already been
            evaluated.
        """
        self.counter += 1
        self.inertia_cnt += 1
        
        for part in self.swarm:
            """ If a new best solution has been found, store it """
            if self.best_scv > G_TOL:
                if part.scv <= G_TOL:
                    self.best_part = part
                    self.best_x = copy.deepcopy(part.x)            
                    self.best_f = copy.deepcopy(part.fval)
                    self.best_scv = copy.deepcopy(part.scv)
                    self.best_part = copy.deepcopy(part)
                    self.counter = 0
                    self.inertia_cnt = 0
            else:
                if part.scv <= G_TOL and part.fval < self.best_f:
                    self.best_part = part
                    self.best_x = copy.deepcopy(part.x)            
                    self.best_f = copy.deepcopy(part.fval)
                    self.best_scv = copy.deepcopy(part.scv)
                    self.best_part = copy.deepcopy(part)
                    self.counter = 0
                    self.inertia_cnt = 0

        # This makes sure iteration can begin even though
        # no feasible particle is found
        if self.best_x is None:
            SCV = [p.scv for p in self.swarm]
            best_idx = np.argmin(SCV)
            self.best_x = self.swarm[best_idx].x
    
    def update_velocity(self,part,verb=False,part_nd=0):
        """ Apply the velocity update rule for particle 'part' """
        rng = np.random.default_rng()
        n = self.problem.nvars()
        
        # Generate multipliers for local and global velocities
        C1 = self.c[0]*rng.random()
        C2 = self.c[1]*rng.random()

        # Store values for the particle
        part.c1 = C1
        part.c2 = C2
        part.w = self.inertia        
        
        # Velocity update
        v0 = self.inertia*part.v + C1*(part.best-part.x) + C2*(self.best_x-part.x)
                        
        #print("v0 = ",v0)
        #print("vmin, vmax = ",self.vmin,self.vmax)
        
        # Keep velocity within velocity bounds
        vnew = np.minimum(np.maximum(v0,self.vmin),self.vmax)
        
        if verb:
            print('---- PARTICLE {0:2g} ----'.format(part_nd))
            print('x = [{0:5.4f},{1:5.4f}]'.format(*part.x))
            print('xbest = [{0:5.4f},{1:5.4f}]'.format(*part.best))
            print('gbest = [{0:5.4f},{1:5.4f}]'.format(*self.best_x))
            print('c1 = {0:5.4f}'.format(part.c1))
            print('c2 = {0:5.4f}'.format(part.c2))
            print('v0 = [{0:5.4f},{1:5.4f}]'.format(*v0))
            print('vnew = [{0:5.4f},{1:5.4f}]'.format(*vnew))
        
        #print("vnew = ",vnew)
        part.v = vnew
    
    def craziness(self,part):
        """ Apply craziness operator """
        rng = np.random.default_rng()
        
        if rng.random() <= self._craziness:
            # change the velocity randomly
            print("Craziness activated")
            part.v = (self.vmax-self.vmin)*rng.random(self.problem.nvars()) + self.vmin
            
    def elite_particle(self):
        """ Replace the worst particle by the best known solution """
        
        
        """ Detect any infeasible points 
            Worst particle is the one with the greatest infeasibility.
            
            NOTE! This has to be modified, if penalty function is used.
        """
        scv = [part.scv for part in self.swarm]
        
        """ Find the index of the particle with greatest constraint violation """
        nd_max = np.argmax(scv)
        if self.swarm[nd_max].scv > G_TOL:
            worst = self.swarm[nd_max]
        else:
            """ If all particles are feasible, find the one with the
                highest objective function value.
            """
            fval = [part.fval for part in self.swarm]
            worst = self.swarm[np.argmax(fval)]
        
        worst = copy.deepcopy(self.best_part)
        
        

    def initizalize_population(self):
        """
        Creates initial population with random positions
        """

        rng = np.random.default_rng()
        n = self.problem.nvars()
        
        xlb = np.array([var.lb for var in self.problem.vars])
        xub = np.array([var.ub for var in self.problem.vars])        
        
        self.vmax = self.vrel*(xub-xlb)
        self.vmin = -self.vmax
        
        for i in range(self.pop_size):
            """ Generate random values for the variables """   
            v0 = (self.vmax-self.vmin)*rng.random(n) + self.vmin
            self.swarm[i] = Particle(self.problem.random_values(),v0)
            self.eval_particle(self.swarm[i])
        
        self.update_best()
        
    def print_pop(self):
        
        print("--- Swarm best ---")
        print("fbest = {0:5.4f}".format(self.best_f))
        if self.best_scv <= G_TOL:
            feas = 'Feasible'
        else:
            feas = 'Infeasible'
            
        print("max sum of constraint violation = {0:5.4f} ({1:s})".format(self.best_scv,feas))
        
            
        print("--- Current Swarm ---")
        for i, part in enumerate(self.swarm):
            print(str(i) + ': ' + repr(part))

    def store_population(self):
        """ Stores the current population """
        
        """
        new_popu = []
        for part in self.swarm:
            new_popu.append(copy.deepcopy(part.x))
        """    
        new_popu = copy.deepcopy(self.swarm)
        self.xvals.append(new_popu)
        self.fvals.append(self.best_x)


    def solve(self, problem, maxiter=100, maxtime=-1, log=False,
              min_diff=1e-5, verb=False, plot=False):

        
        self.problem = problem
        xlb = np.array([var.lb for var in self.problem.vars])
        xub = np.array([var.ub for var in self.problem.vars])
        
        self.penalty = problem.exact_penalty_fun()
        """ Creates the fitness function 
            but the create_fitness method does not return anything!
        """
        self.initizalize_population()
        self.store_population()
        if verb:
            self.print_pop()
        """
        if plot and not self.plot:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            plt.ion()
            plt.show()
            self.plot = True
            self.fig = fig
            self.ax = ax
        """
        
        # Main loop
        it = 0
        prev_val = None
        while it < maxiter:
            it += 1
            problem.num_iters += 1
            print(f"Iteration {it}/{maxiter}")            
            
            """ Perform iteration for particles """
            for i, part in enumerate(self.swarm):
                # Update velocity
                self.update_velocity(part,verb=verb,part_nd=i)
                # Perform craziness
                self.craziness(part)
                # Update design
                # NOTE! This must be modified for discrete variables!
                # (Employ rounding of discrete variables)
                part.move(xlb,xub, self.problem)
                # evaluate particle at its new location
                self.eval_particle(part)
            
            self.update_best()
            self.elite_particle()
            
            self.store_population()
            if verb:
                self.print_pop()
            
            if plot:
                self.update_plot(self.fig, self.ax)

            if self.counter > self.NMAX:
                print(f"The result has not improved in the last 10 generations."
                      f"\n Returning current best values.")
                return self.best_f, self.best_x

            if self.inertia_cnt > self.KMAX:
                print("Reduce inertia and max velocity")
                self.inertia *= 0.9
                self.vmax *= self.vred
                self.vmin *= self.vred

            """
            if log:
                problem.num_iters += 1
                problem.fvals.append(self.best_f)
                problem.states.append(list(pop[np.argmin(fits)]))
                problem.gvals.append(max(self.constr_vals))
                self.fvals.append(self.best_f)
                self.xvals.append(self.best_x)
            """
        # best_idx = np.argmin(fits)
        # xopt = pop[best_idx]
        # print("XOPT: ", xopt)
        # fopt = fits[best_idx]

        return self.best_f, self.best_x

class Particle:
    """ Class for storing a single particle """

    def __init__(self, x, v=None):
        """ Constructor 
            :param x: value for design variables
            :param v: value for velocity
            :param best: best design the particle has visited (personal best)
            :param best_f: objective function value at 'best'
            :param best_scv: sum of constraint violations at 'best'
            :param scv: sum of constraint violations at current location
            :param fval: objective function value at current location
            :param c1, c2: multipliers for local and global velocity components
            :param w: inertia term
            
        """
        self.x = x
        if v is None:
            self.v = np.zeros(len(x))
        else:
            self.v = v
            
        self.best = copy.deepcopy(x)
        self.best_f = np.inf
        self.best_scv = np.inf
        # Sum of constraint violations
        self.scv = np.inf
        # Objective function value
        self.fval = np.inf
        
        self.c1 = None
        self.c2 = None
        self.w = None

    def move(self, xlb, xub, problem):
        """
        Moves particle
        problem: type == OptimizationProblem
        """
        xnew = self.x + self.v

        # Ensure that the particle does not exit the variable bounds
        # (Perhaps the velocity should be changed instead?)

        self.x = np.minimum(np.maximum(xnew,xlb),xub)
        type_of_problem = problem.type_of_vars()

        if type_of_problem == "DiscreteVariable":
            possible_values = problem.vars[0].values
            # Round values
            self.x = np.array([min(possible_values, key=lambda x: abs(x - var))
                      for var in self.x])
    
    def update_best(self):
        """ Updates the best value found so far """
        self.best_scv = copy.deepcopy(self.scv)
        self.best_f = copy.deepcopy(self.fval)
        self.best = copy.deepcopy(self.x)
    
    def __repr__(self):
        s = '\\x &= ['
        for i in range(len(self.x)):
            s += '{' + str(i) + ':5.4f}, '
        
        s = s[:-2]
        
        s += '] & \\vv &= ['
        s = s.format(*self.x)
        
        for i in range(len(self.v)):
            s += '{' + str(i) + ':5.4f}, '
        
        s = s[:-2]
        s += '] & '
        s = s.format(*self.v)
        s += 'SCV &= {0:5.3f} &'.format(self.scv)
        s += 'f &= {0:5.3f}\\\\'.format(self.fval)
        return s


if __name__ == '__main__':
    from metku.optimization.benchmarks import TenBarTruss

    solver = PSO()
    problem = TenBarTruss('continuous')
    fopt, xopt = solver.solve(problem, maxiter=10, verb=False)
    print(fopt, xopt)
    problem(xopt)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

class Support:

    def __init__(self, coordinate, **kwargs):

        values = dict(
            sid=0,
            x=None,
            y=None,
            z=None,
            rx=None,
            ry=None,
            rz=None
        )
        values.update(kwargs)

        self.node = None
        self.node_id = None
        self.coordinate = coordinate
        self.sid = values['sid']





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

from dataclasses import dataclass

import numpy as np

import metku.profiles.steel.isection_formulas as iformulas


@dataclass
class ISection:
    name: str = None
    h: float = 0
    b: float = 0
    tw: float = 0
    tf: float = 0
    r: float = 0

    def __str__(self):
        return f"ISections.{self.name}"

    def __repr__(self):
        return f"ISection({self.name}, {self.h}, {self.b}, {self.tw}, {self.tf}, {self.r})"

    def __copy__(self):
        return ISection(self.name, self.h, self.b, self.tw, self.tf, self.r)

    @property
    def dimensions(self):
        return self.h, self.b, self.tf, self.tw, self.r

    @property
    def A(self):
        return iformulas.area(*self.dimensions)

    @property
    def Avy(self):
        return iformulas.shear_area_y(*self.dimensions)

    @property
    def Avz(self):
        return iformulas.shear_area_z(*self.dimensions)

    @property
    def P(self):
        return iformulas.perimeter(*self.dimensions)

    @property
    def Iy(self):
        return iformulas.second_moment_of_area_y(*self.dimensions)

    @property
    def Iz(self):
        return iformulas.second_moment_of_area_z(*self.dimensions)

    @property
    def I(self):
        return np.array([self.Iy, self.Iz])

    @property
    def It(self):
        return iformulas.torsion_constant(*self.dimensions)

    @property
    def Iw(self):
        return iformulas.warping_constant(*self.dimensions)

    @property
    def Wply(self):
        return iformulas.plastic_section_modulus_y(*self.dimensions)

    @property
    def Wely(self):
        return iformulas.elastic_section_modulus_y(*self.dimensions)

    @property
    def Wplz(self):
        return iformulas.plastic_section_modulus_z(*self.dimensions)

    @property
    def Welz(self):
        return iformulas.elastic_section_modulus_z(*self.dimensions)

    @property
    def iy(self):
        return np.sqrt(self.Iy / self.A)

    @property
    def iz(self):
        return np.sqrt(self.Iz / self.A)


class ISections(ISection):
    # section properties according to:
    # https://eurocodeapplied.com/design/en1993/ipe-hea-heb-hem-design-properties

    # ----------- IPE -------------
    IPE80: ISection = ISection(name="IPE80", h=80.0, b=46.0, tw=3.8, tf=5.2, r=5.0)
    IPE100: ISection = ISection(name="IPE100", h=100.0, b=55.0, tw=4.1, tf=5.7, r=7.0)
    IPE120: ISection = ISection(name="IPE120", h=120.0, b=64.0, tw=4.4, tf=6.3, r=7.0)
    IPE140: ISection = ISection(name="IPE140", h=140.0, b=73.0, tw=4.7, tf=6.9, r=7.0)
    IPE160: ISection = ISection(name="IPE160", h=160.0, b=82.0, tw=5.0, tf=7.4, r=9.0)
    IPE180: ISection = ISection(name="IPE180", h=180.0, b=91.0, tw=5.3, tf=8.0, r=9.0)
    IPE200: ISection = ISection(name="IPE200", h=200.0, b=100.0, tw=5.6, tf=8.5, r=12.0)
    IPE220: ISection = ISection(name="IPE220", h=220.0, b=110.0, tw=5.9, tf=9.2, r=12.0)
    IPE240: ISection = ISection(name="IPE240", h=240.0, b=120.0, tw=6.2, tf=9.8, r=15.0)
    IPE270: ISection = ISection(name="IPE270", h=270.0, b=135.0, tw=6.6, tf=10.2, r=15.0)
    IPE300: ISection = ISection(name="IPE300", h=300.0, b=150.0, tw=7.1, tf=10.7, r=15.0)
    IPE330: ISection = ISection(name="IPE330", h=330.0, b=160.0, tw=7.5, tf=11.5, r=18.0)
    IPE360: ISection = ISection(name="IPE360", h=360.0, b=170.0, tw=8.0, tf=12.7, r=18.0)
    IPE400: ISection = ISection(name="IPE400", h=400.0, b=180.0, tw=8.6, tf=13.5, r=21.0)
    IPE450: ISection = ISection(name="IPE450", h=450.0, b=190.0, tw=9.4, tf=14.6, r=21.0)
    IPE500: ISection = ISection(name="IPE500", h=500.0, b=200.0, tw=10.2, tf=16.0, r=21.0)
    IPE550: ISection = ISection(name="IPE550", h=550.0, b=210.0, tw=11.1, tf=17.2, r=24.0)
    IPE600: ISection = ISection(name="IPE600", h=600.0, b=220.0, tw=12.0, tf=19.0, r=24.0)
    # ----------- HEA -------------
    HEA100: ISection = ISection(name="HEA100", h=96.0, b=100.0, tw=5.0, tf=8.0, r=12.0)
    HEA120: ISection = ISection(name="HEA120", h=114.0, b=120.0, tw=5.0, tf=8.0, r=12.0)
    HEA140: ISection = ISection(name="HEA140", h=133.0, b=140.0, tw=5.5, tf=8.5, r=12.0)
    HEA160: ISection = ISection(name="HEA160", h=152.0, b=160.0, tw=6.0, tf=9.0, r=15.0)
    HEA180: ISection = ISection(name="HEA180", h=171.0, b=180.0, tw=6.0, tf=9.5, r=15.0)
    HEA200: ISection = ISection(name="HEA200", h=190.0, b=200.0, tw=6.5, tf=10.0, r=18.0)
    HEA220: ISection = ISection(name="HEA220", h=210.0, b=220.0, tw=7.0, tf=11.0, r=18.0)
    HEA240: ISection = ISection(name="HEA240", h=230.0, b=240.0, tw=7.5, tf=12.0, r=21.0)
    HEA260: ISection = ISection(name="HEA260", h=250.0, b=260.0, tw=7.5, tf=12.5, r=24.0)
    HEA280: ISection = ISection(name="HEA280", h=270.0, b=280.0, tw=8.0, tf=13.0, r=24.0)
    HEA300: ISection = ISection(name="HEA300", h=290.0, b=300.0, tw=8.5, tf=14.0, r=27.0)
    HEA320: ISection = ISection(name="HEA320", h=310.0, b=300.0, tw=9.0, tf=15.5, r=27.0)
    HEA340: ISection = ISection(name="HEA340", h=330.0, b=300.0, tw=9.5, tf=16.5, r=27.0)
    HEA360: ISection = ISection(name="HEA360", h=350.0, b=300.0, tw=10.0, tf=17.5, r=27.0)
    HEA400: ISection = ISection(name="HEA400", h=390.0, b=300.0, tw=11.0, tf=19.0, r=27.0)
    HEA450: ISection = ISection(name="HEA450", h=440.0, b=300.0, tw=11.5, tf=21.0, r=27.0)
    HEA500: ISection = ISection(name="HEA500", h=490.0, b=300.0, tw=12.0, tf=23.0, r=27.0)
    HEA550: ISection = ISection(name="HEA550", h=540.0, b=300.0, tw=12.5, tf=24.0, r=27.0)
    HEA600: ISection = ISection(name="HEA600", h=590.0, b=300.0, tw=13.0, tf=25.0, r=27.0)
    HEA650: ISection = ISection(name="HEA650", h=640.0, b=300.0, tw=13.5, tf=26.0, r=27.0)
    HEA700: ISection = ISection(name="HEA700", h=690.0, b=300.0, tw=14.5, tf=27.0, r=27.0)
    HEA800: ISection = ISection(name="HEA800", h=790.0, b=300.0, tw=15.0, tf=28.0, r=30.0)
    HEA900: ISection = ISection(name="HEA900", h=890.0, b=300.0, tw=16.0, tf=30.0, r=30.0)
    HEA1000: ISection = ISection(name="HEA1000", h=990.0, b=300.0, tw=16.5, tf=31.0, r=30.0)
    # ----------- HEB -------------
    HEB100: ISection = ISection(name="HEB100", h=100.0, b=100.0, tw=6.0, tf=10.0, r=12.0)
    HEB120: ISection = ISection(name="HEB120", h=120.0, b=120.0, tw=6.5, tf=11.0, r=12.0)
    HEB140: ISection = ISection(name="HEB140", h=140.0, b=140.0, tw=7.0, tf=12.0, r=12.0)
    HEB160: ISection = ISection(name="HEB160", h=160.0, b=160.0, tw=8.0, tf=13.0, r=15.0)
    HEB180: ISection = ISection(name="HEB180", h=180.0, b=180.0, tw=8.5, tf=14.0, r=15.0)
    HEB200: ISection = ISection(name="HEB200", h=200.0, b=200.0, tw=9.0, tf=15.0, r=18.0)
    HEB220: ISection = ISection(name="HEB220", h=220.0, b=220.0, tw=9.5, tf=16.0, r=18.0)
    HEB240: ISection = ISection(name="HEB240", h=240.0, b=240.0, tw=10.0, tf=17.0, r=21.0)
    HEB260: ISection = ISection(name="HEB260", h=260.0, b=260.0, tw=10.0, tf=17.5, r=24.0)
    HEB280: ISection = ISection(name="HEB280", h=280.0, b=280.0, tw=10.5, tf=18.0, r=24.0)
    HEB300: ISection = ISection(name="HEB300", h=300.0, b=300.0, tw=11.0, tf=19.0, r=27.0)
    HEB320: ISection = ISection(name="HEB320", h=320.0, b=300.0, tw=11.5, tf=20.5, r=27.0)
    HEB340: ISection = ISection(name="HEB340", h=340.0, b=300.0, tw=12.0, tf=21.5, r=27.0)
    HEB360: ISection = ISection(name="HEB360", h=360.0, b=300.0, tw=12.5, tf=22.5, r=27.0)
    HEB400: ISection = ISection(name="HEB400", h=400.0, b=300.0, tw=13.5, tf=24.0, r=27.0)
    HEB450: ISection = ISection(name="HEB450", h=450.0, b=300.0, tw=14.0, tf=26.0, r=27.0)
    HEB500: ISection = ISection(name="HEB500", h=500.0, b=300.0, tw=14.5, tf=28.0, r=27.0)
    HEB550: ISection = ISection(name="HEB550", h=550.0, b=300.0, tw=15.0, tf=29.0, r=27.0)
    HEB600: ISection = ISection(name="HEB600", h=600.0, b=300.0, tw=15.5, tf=30.0, r=27.0)
    HEB650: ISection = ISection(name="HEB650", h=650.0, b=300.0, tw=16.0, tf=31.0, r=27.0)
    HEB700: ISection = ISection(name="HEB700", h=700.0, b=300.0, tw=17.0, tf=32.0, r=27.0)
    HEB800: ISection = ISection(name="HEB800", h=800.0, b=300.0, tw=17.5, tf=33.0, r=30.0)
    HEB900: ISection = ISection(name="HEB900", h=900.0, b=300.0, tw=18.5, tf=35.0, r=30.0)
    HEB1000: ISection = ISection(name="HEB1000", h=1000.0, b=300.0, tw=19.0, tf=36.0, r=30.0)
    # ----------- HEM -------------
    HEM100: ISection = ISection(name="HEM100", h=120.0, b=106.0, tw=12.0, tf=20.0, r=12.0)
    HEM120: ISection = ISection(name="HEM120", h=140.0, b=126.0, tw=12.5, tf=21.0, r=12.0)
    HEM140: ISection = ISection(name="HEM140", h=160.0, b=146.0, tw=13.0, tf=22.0, r=12.0)
    HEM160: ISection = ISection(name="HEM160", h=180.0, b=166.0, tw=14.0, tf=23.0, r=15.0)
    HEM180: ISection = ISection(name="HEM180", h=200.0, b=186.0, tw=14.5, tf=24.0, r=15.0)
    HEM200: ISection = ISection(name="HEM200", h=220.0, b=206.0, tw=15.0, tf=25.0, r=18.0)
    HEM220: ISection = ISection(name="HEM220", h=240.0, b=226.0, tw=15.5, tf=26.0, r=18.0)
    HEM240: ISection = ISection(name="HEM240", h=270.0, b=248.0, tw=18.0, tf=32.0, r=21.0)
    HEM260: ISection = ISection(name="HEM260", h=290.0, b=268.0, tw=18.0, tf=32.5, r=24.0)
    HEM280: ISection = ISection(name="HEM280", h=310.0, b=288.0, tw=18.5, tf=33.0, r=24.0)
    HEM300: ISection = ISection(name="HEM300", h=340.0, b=310.0, tw=21.0, tf=39.0, r=27.0)
    HEM320: ISection = ISection(name="HEM320", h=359.0, b=309.0, tw=21.0, tf=40.0, r=27.0)
    HEM340: ISection = ISection(name="HEM340", h=377.0, b=309.0, tw=21.0, tf=40.0, r=27.0)
    HEM360: ISection = ISection(name="HEM360", h=395.0, b=308.0, tw=21.0, tf=40.0, r=27.0)
    HEM400: ISection = ISection(name="HEM400", h=432.0, b=307.0, tw=21.0, tf=40.0, r=27.0)
    HEM450: ISection = ISection(name="HEM450", h=478.0, b=307.0, tw=21.0, tf=40.0, r=27.0)
    HEM500: ISection = ISection(name="HEM500", h=524.0, b=306.0, tw=21.0, tf=40.0, r=27.0)
    HEM550: ISection = ISection(name="HEM550", h=572.0, b=306.0, tw=21.0, tf=40.0, r=27.0)
    HEM600: ISection = ISection(name="HEM600", h=620.0, b=305.0, tw=21.0, tf=40.0, r=27.0)
    HEM650: ISection = ISection(name="HEM650", h=668.0, b=305.0, tw=21.0, tf=40.0, r=27.0)
    HEM700: ISection = ISection(name="HEM700", h=716.0, b=304.0, tw=21.0, tf=40.0, r=27.0)
    HEM800: ISection = ISection(name="HEM800", h=814.0, b=303.0, tw=21.0, tf=40.0, r=30.0)
    HEM900: ISection = ISection(name="HEM900", h=910.0, b=302.0, tw=21.0, tf=40.0, r=30.0)
    HEM1000: ISection = ISection(name="HEM1000", h=1008.0, b=302.0, tw=21.0, tf=40.0, r=30.0)


if __name__ == '__main__':
    ipe = ISections.IPE270

    print(ipe)



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
""" Class for steel members """

import math

import numpy as np

    
from metku.eurocodes.en1993 import constants, en1993_1_1
from metku.sections.steel.wi import WISection


class WebTaperedSteelMember:
    """ Web tapered steel members accroding to EN 1993                     
        TODO:
            - NbRd
            - MbRd            
            - Nurjahdus
            - Kiepahdus

    """

    def __init__(self, profile0, h1, length, Lcr=[1.0, 1.0], mtype="beam",
                 LT_buckling=False,taper_type="right"):
        """ Constructor
            
            profile0 -- smaller member profile
            h1 -- height of the profile at the other end
            length -- member length
            lcr -- array with length 2 including the buckling
                    length factors
            type -- member type (beam, column or truss)
            taper_type .. type of tapering
                        "right" right hand side of the profile is tapered
                        "left" left hand side of the profile is tapered
                        "dual" both sides are tapered
            
            Other attributes:
            ned -- axial force in various sections (array)
            myed -- bending moment in various sections 
                    with respect to major axis (array)
            mzed -- bending moment in various sections
                    with respect to minor axis (array)
            vyed -- shear force in various sections
                    with respect to major axis (array)
            vzed -- shear force in various sections
                    with respect to minor axis (array)
            ted -- torsion moment in various sections (array)
            loc -- locations of sections along the members in
                    local coordinates (0 <= loc[i] <= 1)
            
        """
        
        # create 'top profile'
        profile1 = WISection(h1,profile0.tw,profile0.b,profile0.tf,profile0.fy,profile0.weld_throat)

        self.profiles = [profile0,profile1]
        self.length = length
        self.lcr = np.array(Lcr)
        self.h1 = h1
        # self.lcr[0] = Lcr[0]*length
        # self.lcr[1] = Lcr[1]*length
        self.type = mtype
        self.ned = []
        self.myed = []
        self.mzed = []
        self.vyed = []
        self.vzed = []
        self.ted = []
        self.loc = []
        self.LT_B = LT_buckling
        self.taper_type = taper_type

    @property
    def NbRd(self):
        return np.asarray(self.buckling_strength())

    @property
    def MbRd(self):
        Mcr = self.mcrit()
        return np.asarray(self.LT_buckling_strength(Mcr))
    
    @property
    def MEdY(self):
        return max(abs(np.array(self.myed)))
    
    @property
    def MEdZ(self):
        return max(abs(np.array(self.mzed)))

    @property
    def NEd(self):
        return max(abs(np.array(self.ned)))

    @property
    def Iymax(self):
        return self.profiles[1].I[0]

    @property
    def Iymin(self):
        return self.profiles[0].I[0]

    @property
    def Ieq(self):
        """ Equivalent second moment of area with respect to major axis """
        
        Iymax = self.Iymax
        r = math.sqrt(self.Iymin/Iymax)
        C = 0.08+0.92*r
        
        return C*Iymax

    def nsect(self):
        """ Number of sections """
        return len(self.ned)

    def fy(self):
        """ Yield strength of the member """
        return self.profile.fy

    def ncrit(self, verb=False):
        """ Buckling force according to Euler """
        C = math.pi ** 2 * self.profile.E

        ncrit = C * np.array(self.profile.I) / (self.length * self.lcr) ** 2

        if verb:
            print("Ncr,y = {0:4.2f} kN".format(ncrit[0] * 1e-3))
            print("Ncr,z = {0:4.2f} kN".format(ncrit[1] * 1e-3))

        return ncrit

    def slenderness(self, verb=False):
        """ Non-dimensional slenderness according to EN 1993-1-1 """
        NRd = self.profile.A * self.fy()
        Ncr = self.ncrit(verb)
        # if NRd <= 1e-6:
        # print(self.profile.A)
        # print(self.profile.h)
        # print(NRd, Ncr)

        slend = np.sqrt(NRd / Ncr)

        if verb:
            print("lambda,y = {0:4.2f}".format(slend[0]))
            print("lambda,z = {0:4.2f}".format(slend[1]))

        return slend

    def LT_slenderness(self, Mcr, verb=False):
        """ Non-dimensional slenderness for lateral-torsional buckling.
        """
        MRd = self.profile.MRd
        lambdaLT = np.sqrt(MRd[0] / Mcr)
        if verb:
            # print("MRd: {}, Mcr: {}".format(MRd, Mcr))
            # print("MRd0 = {0:4.2f}".format(MRd[0] * 1e-6))
            if MRd[0] < 0:
                print(self.profile.h, self.profile.tw, self.profile.tf, self.profile.b)
            # print("MRd1 = {0:4.2f}".format(MRd[1] * 1e-6))
            # print("Mcr = {0:4.2f}".format(Mcr * 1e-6))
        return lambdaLT

    def buckling_strength(self, verb=False):
        """ Member buckling strength according to EN 1993-1-1 """

        if verb:
            print("** Buckling strength** ")

        slend = self.slenderness(verb)
        NbRd = []
        NRd = self.profile.A * self.fy()
        # self.profile.imp_factor()
        # print(slend)
        # p = 0.5*(1+np.array(self.profile.imp_factor)*(slend-0.2)+slend**2)
        # r = 1/(p+np.sqrt(p**2-slend**2))

        for i in range(len(slend)):
            p = 0.5 * (1 + self.profile.imp_factor[i] * (slend[i] - 0.2) +
                       slend[i] ** 2)
            r = min(1 / (p + math.sqrt(p ** 2 - slend[i] ** 2)), 1.0)
            if verb:
                print("Psi,{0:1} = {1:4.2f}".format(i, p))
                print("chi,{0:1} = {1:4.2f}".format(i, r))
                print("NbRd,{0:1} = {1:4.2f}".format(i, NRd * r * 1e-3))

            NbRd.append(NRd * r)

        # NbRd = self.profile.A*self.fy()*r;
        return NbRd

    def LT_buckling_strength(self, Mcr, axis='y', method='specific',
                             verb=False):
        """ Member lateral-torsional bending strength """

        MRd = self.profile.bending_resistance()[0]
        if axis == 'y':
            idx = 0
        else:
            idx = 1
        lambdaLT = self.LT_slenderness(Mcr) # [idx]
        
        # self.profile.define_imp_factor_LT()
        # alpha = self.profile.ImperfectionLT

        if method == 'general':
            alphaLT = self.profile.imp_factor_LT_gen
            lambdaLT0 = 0.2
            beta = 1.0
            chiLTmax = 1.0
        else:
            alphaLT = self.profile.imp_factor_LT
            lambdaLT0 = 0.4
            beta = 0.75
            chiLTmax = min(1.0, 1/lambdaLT**2)
        
        if lambdaLT <= lambdaLT0 or self.MEdY/Mcr <= lambdaLT0**2:
            chiLT = 1.0
        else:
            p = 0.5 * (1 + alphaLT * (lambdaLT - lambdaLT0) + beta*lambdaLT ** 2)
            chiLT = min(1. / (p + math.sqrt(p ** 2 - beta*lambdaLT ** 2)), chiLTmax
                        )
        MbRd = chiLT * MRd
        
        if verb:
            print("lambdaLT = {0:4.2f}".format(lambdaLT))
            print("alphaLT = {0:4.2f}".format(alphaLT))
            # print("phi = {0:4.2f}".format(p))
            print("chi_LT = {0:4.2f}".format(chiLT))
            print("MbRd = {0:4.2f}".format(MbRd*1e-6))
            print("MRd = {0:4.2f}".format(MRd * 1e-6))

        return MbRd

    def weight_per_length(self):
        """ Weight per unit length
            
            Default units: kg/mm
        """
        w = self.profile.A * constants.density
        return w

    def weight(self):
        """ Weight of the member (kg) """
        w = self.weight_per_length * self.length
        return w

    def mcrit(self, C=[1.12, 0.45, 0.525], k=[1, 1]):
        """ Critical moment for lateral-torsional buckling

            For mono- and double-symmetric profiles
            C(1), C(2), C(3) -- coefficients
            za -- position of application of load
            zs -- position of the shear centre S from centroid G
            zj -- ??
            k(1) -- kz
            k(2) -- kw
        """

        kz = k[0]
        kw = k[1]

        Iz = self.profile.I[1]
        Iw = self.profile.Iw
        It = self.profile.It

        E = constants.E
        G = constants.G
        L = self.length
        C1, C2, C3 = C

        if self.symmetry == 'dual':
            za = 0
            zs = 0
            zg = za - zs
            part1 = C1 * (math.pi ** 2 * E * Iz) / (kz * L) ** 2
            part2 = np.sqrt((kz / kw) ** 2 * Iw / Iz + (kz * L) ** 2 * G * It /
                            (math.pi ** 2 * E * Iz) + (C2 * zg) ** 2)
            part3 = C2 * zg

            Mcr = part1 * (part2 - part3)

        elif self.symmetry == 'mono':
            za = 0
            zs = self.profile.zs
            zg = za - zs

            zj = self.profile.zj

            part1 = C1 * (math.pi ** 2 * E * Iz) / (kz * L) ** 2
            part2 = C3 * zj - C2 * zg
            part3 = np.sqrt((kz / kw) ** 2 * Iw / Iz + (kz * L) ** 2 * G * It /
                            (math.pi ** 2 * E * Iz) + (C2 * zg - C3 * zj) ** 2)

            Mcr = part1 * (part2 + part3)

        # print("Iz = {0:4.2f}".format(Iz*1e-4))
        # print("Iw = {0:4.2f}".format(Iw*1e-6))
        # print("It = {0:4.2f}".format(It*1e-4))
        # print("Mcr = {0:4.2f}".format(Mcr*1e-6))
        # print("kz =", kz, "kw =", kw, "L =", L, "part3 =", part3))

        return Mcr

    def add_section(self, ned=0.0, myed=0.0, mzed=0.0, vyed=0.0, vzed=0.0,
                    ted=0.0, loc=0.0):
        """ Adds new section with internal forces and location """

        self.ned.append(ned)
        self.myed.append(myed)
        self.mzed.append(mzed)
        self.vyed.append(vyed)
        self.vzed.append(vzed)
        self.ted.append(ted)
        self.loc.append(loc)

    def check_section(self, n=0):
        """ Verify resistance of section 'n' """
        self.profile.Ned = self.ned[n]
        self.profile.Med = self.myed[n]
        self.profile.Ved = self.vzed[n]

        r = self.profile.section_resistance()
        return r

    def check_sections(self, class1or2=True):
        """ Verify resistance of all sections """
        r = np.zeros(len(self.check_section()))
        for n in range(self.nsect()):
            r = np.vstack((r, self.check_section(n)))
            # r.append(self.check_section(n))
        return np.max(r, axis=0)

    def check_buckling(self):
        """ Verify buckling resistance
            output: r .. utilization ratio for buckling        
        """

        """ Find greatest compression force in the member
            clip replace all positive values with 0.0        
        """
        NEd = np.min(np.array(self.ned).clip(max=0.0))
        if NEd >= 0.0:
            """ if all axial forces are non-negative,
                buckling resistance need not be checked
            """
            r = [0.0, 0.0]
        else:
            r = abs(NEd) / self.NbRd

        return r

    def check_LT_buckling(self):
        """ Verify lateral torsional buckling resistance
            output: r .. utilization ratio for buckling        
        """

        MEd1 = np.max(np.array(self.myed))
        MEd2 = abs(np.min(np.array(self.myed)))
        MEd = max(MEd1, MEd2)
        MbRd = self.LT_buckling_strength(self.mcrit())
        r = abs(MEd) / MbRd

        return r

    def check_beamcolumn(self, Cmy=0.9, section_class=None):
        """ Verify stability for combined axial force and bending moment
            LT_buckling -- kiepahdus
                    True .. rajoitusehto huomioidaan
                    False .. rajoitusehtoa ei huomioida
        """
        # find the largest compression force in the member
        NEd = np.min(np.array(self.ned).clip(max=0.0))

        self.profile.Ned = NEd
        if section_class is None:
            cross_section_class = self.profile.section_class()
        else:
            cross_section_class = section_class
            
        #print("Check beam column:")
        #print("Section class:", cross_section_class)
        
        slend = self.slenderness()
        CmLT = 1  # Pitääkö käyttää jotakin kaavaa?

        # NRd = self.profile.A * self.fy()
        NbRd = self.buckling_strength()

        UNy = abs(NEd) / NbRd[0]
        UNz = abs(NEd) / NbRd[1]

        kyy = en1993_1_1.kyy(UNy, slend[0], Cmy,
                             section_class=cross_section_class)

        kzy = en1993_1_1.kzy(kyy, UNz, slend[1], CmLT=CmLT,
                             section_class=cross_section_class)

        MyEd1 = np.max(np.array(self.myed))
        MyEd2 = abs(np.min(np.array(self.myed)))
        MyEd = max(MyEd1, MyEd2)

        MRd = self.profile.bending_resistance(C=cross_section_class)    

        com_comp_bend_y = -NEd / NbRd[0] + kyy * MyEd / MRd[0]
        com_comp_bend_z = -NEd / NbRd[1] + kzy * MyEd / MRd[0]

        com_comp_bend = [com_comp_bend_y, com_comp_bend_z]

        return com_comp_bend

if __name__ == '__main__':
    
    p = WISection(200, 6.0, [140.0, 140.0], [8.0, 8.0],fy=355,weld_throat=4)
    
    m = WebTaperedSteelMember(p, 400, 8000, Lcr=[1.0, 1.0], mtype="column",LT_buckling=True,taper_type="right")
    
    print(m.Iymin,m.Ieq,m.Iymax)    
    


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 19:34:15 2021

Cold-formed trapezoidal sheeting profiles

@author: kmela
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
from metku.materials.steel_data import Steel
from metku.eurocodes.en1993.en1993_1_3.open_prof import OpenProf
from metku.eurocodes.en1993.en1993_1_3 import en1993_1_3
from metku.eurocodes.en1993 import en1993_1_5
from metku.eurocodes.en1993.constants import gammaM0, gammaM1


class TrapezoidalSheet:
    """ Class for trapezoidal sheeting """
    
    def __init__(self,t,h,bt,bb,angle,r=3,material=Steel("S350GD"),ntroughs=3):
        """
        Constructor

        Parameters
        ----------
        t : float
            Nominal thickness.
        h : float
            height of profile.
        bt : float
            width of top flange.
        bb : float
            width of bottom flange.
        angle : float
            angle between bottom flange and chord (degrees).
        r : float
            rounding (interal)
        material : Steel, optional
            material. The default is Steel("350GD").
        ntroughs : float
            number of troughs in one sheet

        Returns
        -------
        None.

        """
        
        self.tnom = t
        self.tcoat = 0.04
        self.t = self.tnom-self.tcoat
        self.h = h
        self.bt = bt
        self.bb = bb
        self.angle = angle
        self.angle_rad = np.radians(angle)
        self.material = material
        self.ntrough = ntroughs
        
        
    @property
    def E(self):
        return self.material.E
    
    @property
    def nu(self):
        return self.material.nu
    
    @property
    def fy(self):
        return self.material.fy
    
    @property
    def fu(self):
        return self.material.fu
    
    @property
    def hw(self):
        """ Profile height between center lines """
        return self.h-self.tnom
    
    @property
    def sw(self):
        """ Length of one web """
        return self.hw*np.sin(self.angle_rad)
    
    @property
    def pitch(self):
        """ Pitch of corrugations """
        return self.bb+self.bt+2*self.h*np.cos(self.angle_rad)
    
    @property
    def width(self):
        """ Sheet width """
        return self.pitch*self.ntrough
    
    @property
    def A(self):
        """ Cross-sectional area """
        return self.ntrough*self.t*(self.bt+self.bb+2*self.sw)
    


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Sun Jun 28 15:02:50 2020

@author: kmela
"""



from metku.frame2d.frame2d import Frame2D, FrameMember, PointLoad, LineLoad, \
    Support, Hinge, XYHingedSupport, PREC
from metku.framefem.elements import EBBeam, EBSemiRigidBeam
# from fem.elements.eb_semi_rigid_beam import EBSemiRigidBeam
from metku.sections.steel import HEA

from metku.framefem import FrameFEM, BeamSection
from metku.structures.steel.steel_member import SteelMember
from metku.eurocodes.en1993.en1993_1_8.rhs_joints import RHSKGapJoint, \
    RHSYJoint

    
import numpy as np
from timeit import default_timer as timer

class SimpleTruss(Frame2D):
    """ Class for simple trusses. Essentially, the structure is a collection of
        pin-jointed members, loaded at the nodes. Consequently, the members
        carry only axial forces.
        
        This class is mainly used for truss optimization problems.
        
        Methods:
        ------------
        add
        generate
        calculate
        plot
    """
    
    def __init__(self,nodes=None,members=None,profiles=None):
        """ Constructor
        

        Parameters:
        ------------
        :param nodes: list of nodal coordinates [[x1,y1],[x2,y2],...,[xn,yn]]
        :param members: list of member nodes [[e11,e12],[e21,e22],...,[em1,em2]]
            where eij is a node index in the list 'nodes'
        :param profiles: list of member profiles, corresponding to the members

        :type nodes: list of lists
        :type members: list of lists
        :type profile: list


        Variables:
        ----------
        :ivar nodes:
        
        """
        super().__init__(num_elements=1)
    
        self._nodes = nodes
        self._members = members
        self._profiles = profiles
        
        """ Create truss members """
        for mem, prof in zip(members,profiles):
            mem_nodes = [nodes[mem[0]],nodes[mem[1]]]
            self.add(SimpleTrussMember(mem_nodes,profile=prof))
    
    def add(self, this):
        """
        Adds object to truss
        :param this: object to be added
        :return:
        """
        # Simple truss member
        if isinstance(this, SimpleTrussMember):
            this.mem_id = int(len(self.members))
            self.members[this.mem_id] = this
            this.calc_nodal_coordinates(self.num_elements)            
        else:
            super().add(this)
    #def generate(self):
    #    """ Generates the finite element model """ 
        
    
    
class SimpleTrussMember(FrameMember):
    """ Class for simple truss members. These are a subclass of FrameMember,
        but with less functionalities, because only axial forces are carried.
        
        On the other hand, the profile can be generic, i.e. only 'A' and optionally 'I'
        can be given.
    """
    
    def __init__(self,
                 coordinates,
                 mem_id="",
                 profile="SHS 100x100x5.0",
                 material="S355",
                 mtype='bar',
                 num_elements=1):

        super().__init__(coordinates, mem_id, profile, material, num_elements,mtype=mtype)

def three_bar_truss(L=1000,F1=100,F2=100,nlc=1):
    """ three bar truss example """
    
    #from metku.sections.cross_sections import CrossSection

    # Nodal coordinates    
    X = [[0,0],[-L,L],[0,L],[round(np.sqrt(3)*L,PREC),L]]
    
    """ Member connectivities. For example, [0,1] means that the member
        end nodes are the 0th and 1st nodes in the list.
    """
    mems = [[0,1],[0,2],[0,3]]
    
    #profs = [SHS(50,5),SHS(50,5),SHS(50,5)]
    """ Cross-sections
        Here, a name for the cross-section has to be given. This has to do
        with the way FrameMember has been implemented.
        
        Later, the profile data can be changed manually, e.g. by modifying
        the cross-sectional area 'A' directly.
    """
    profs = ['SHS 50x50x3.0','SHS 50x50x3.0','SHS 50x50x3.0']
    #profs = [CrossSection(A=100),CrossSection(A=100),CrossSection(A=100)]

    # Generate truss
    t = SimpleTruss(X,mems,profs)
    
    # Add hinged supports
    t.add(XYHingedSupport(X[1]))
    t.add(XYHingedSupport(X[2]))
    t.add(XYHingedSupport(X[3]))
    # Add point load
    # Voima F1 on pystysuuntainen ja F2 vaakasuuntainen
    if nlc == 1:
        p = PointLoad(X[0],[F2,F1,0])
        t.add(p)
    else:
        p1 = PointLoad(X[0],[F2,0,0],load_id=2)
        t.add(p1)
        p2 = PointLoad(X[0],[0,F1,0],load_id=3)
        t.add(p2)
    
    #t.plot()
    
    # Generate calculation model
    t.generate()
    # Calculate the responses
    t.calculate(support_method='REM')
    
    return t
        
def ten_bar_truss(L,F):
    """ Classic 10-bar truss """
    nodes = []
    nodes.append([2 * L, L])
    nodes.append([2 * L, 0])
    nodes.append([L, L])
    nodes.append([L, 0])
    nodes.append([0, 0])
    nodes.append([0, L])
    
    mems = [[4,2],[2,0],[5,3],[3,1],[2,3],[0,1],[4,3],[5,2],[2,1],[3,0]]
    
    profs = []
    for i in range(10):
        profs.append('SHS 50x50x3.0')
    
    t = SimpleTruss(nodes,mems,profs)
    
    t.add(PointLoad(nodes[1],[0,-F,0]))
    t.add(PointLoad(nodes[3],[0,-F,0]))
    
    t.add(XYHingedSupport(nodes[4]))
    t.add(XYHingedSupport(nodes[5]))
    
    t.generate()
    t.f.nodal_dofs()
    #t.calculate()
    
    #t.plot()
    return t

def fifteen_bar_truss(L=360,h=120,P=10000):
    """ by default, the units are imperial """
    
    nodes = []
    nodes.append([0, h])
    nodes.append([L/3, h])
    nodes.append([2*L/3, h])
    nodes.append([L, h])
    nodes.append([0, 0])
    nodes.append([L/3, 0])
    nodes.append([2*L/3, 0])
    nodes.append([L, 0])
    
    mems = [[0,1],[1,2],[2,3],[4,5],[5,6],[6,7],
            [5,1],[6,2],[7,3],
            [0,5],[4,1],
            [1,6],[5,2],
            [2,7],[6,3]]
    
    profs = []
    for i in range(len(mems)):
        profs.append('SHS 50x50x3.0')
    
    t = SimpleTruss(nodes,mems,profs)
        
    t.add(PointLoad(nodes[7],[0,-P,0]))
            
    t.add(XYHingedSupport(nodes[0]))
    t.add(XYHingedSupport(nodes[4]))
    
    t.generate()
    
    #print(type(t.point_loads['PointLoad'].coordinate))
    
    t.calculate(load_id='all',support_method="REM")
    
    t.plot()
    
    return t
    
if __name__ == '__main__':
    
    import cProfile
    #t = three_bar_truss(L=3000,F1=-200e3,F2=-250e3,nlc=1)
    t = ten_bar_truss(L=3000,F=200e3)
    
    #t = fifteen_bar_truss()
    
    t.generate()
    load_id = t.load_ids[0]
    #start = timer()    
    #t.f.linear_statics(lcase=load_id,support_method="REM")
    #t.calculate(load_id="all",support_method="REM")
    #end = timer()
    #print(end-start)
    cProfile.run('t.f.linear_statics(lcase=load_id,support_method="REM")')
    #cProfile.run('t.calculate(load_id="all",support_method="REM")')
        
       


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
'''
Ant colony optimization for truss structures.
Author: Olli Luukkonen

Limitations:
- Supports only truss structures.
- Works only if variable type is discrete. Does NOT work for continuous or index variables.
- Results are not perfect, for example pso.py gives better results than this algorithm.
'''


from metku.optimization.solvers import OptSolver
from metku.frame2d import simple_truss
from metku.optimization.structopt import Variable, OptimizationProblem, \
    IndexVariable, DiscreteVariable, IntegerVariable
from metku.optimization.solvers.optsolver import OptSolver, G_TOL

import numpy as np
import warnings

warnings.filterwarnings("ignore")

class AntColonyOptimizer(OptSolver):
    def __init__(self, pop_size=100, rho=1, alpha=1,
                                beta=0.5, xi=0.75, number_of_top_ants=4,
                                penalty_factor=10000):

        super().__init__()

        self.number_of_ants = pop_size

        # Factor for global pheromone update
        self.rho = rho

        self.penalty = None

        self.pheromone_matrix = None
        self.heuristic_matrix = None
        self.probability_matrix = None

        self.alpha = alpha
        self.beta = beta

        # Factor for local pheromone update
        self.xi = xi

        self.number_of_top_ants = number_of_top_ants

        # Factor which increases penalty value
        self.penalty_factor = penalty_factor

        self.design_variables = []
        self.best_fitness = np.inf
        self.best_x = None
        self.min_f = None

    def get_min_f(self):
        '''
        :return: Minimum objective function value
        '''
        min_f = 0
        for member in self.problem.structure.members.values():
            min_f += self.calculate_member_mass(member, min(self.design_variables))
        return min_f

    def calculate_member_mass(self, member, cross_section_area):
        return member.length * cross_section_area * member.rho

    def initialize_heuristic_matrix(self):
        self.heuristic_matrix = np.full((self.i, self.j), 0, dtype=np.float64)
        i = 0
        for member in self.problem.structure.members.values():
            for j in range(self.j):
                self.heuristic_matrix[i][j] = 1 / self.calculate_member_mass(member, self.design_variables[j])
            i += 1

    def update_probabilities(self):
        self.probability_matrix = (self.pheromone_matrix ** self.alpha) * (
                self.heuristic_matrix ** self.beta)

    def initialize(self):
        '''
        Initialize parameters and matrices
        '''
        self.design_variables = self.problem.vars[0].values
        number_of_members = len(self.problem.vars)

        self.min_f = self.get_min_f()

        self.i = number_of_members
        self.j = len(self.design_variables)

        initial_pheromone = 1 / self.get_min_f()
        self.pheromone_matrix = np.full((self.i, self.j), initial_pheromone, dtype=np.float64)
        self.initialize_heuristic_matrix()
        self.update_probabilities()

    def eval_ant(self, ant, x):
        """
        Evaluates one ant:
        Calculate objective function, penalty value and fitness
        :param ant: Ant class object
        :return: fitness
        """
        self.problem.substitute_variables(x)
        """ Evaluate objective function """
        ant.fval = self.problem.obj(x)
        """ Evaluate penalty function """

        ant.penalty = self.penalty(x) * self.penalty_factor

        """ Fitness function equals objective function value + penalty """
        val = ant.fval + ant.penalty

        return val

    def evaluate(self, ants):
        """
        Calculate fitness value for all ants
        :param ants: List of ant objects
        :return: best objective function, list of best design variables
        """
        best_x = []
        best_fitness = np.inf
        best_ant = None
        for ant in ants:
            x = ant.get_x()
            fitness = self.eval_ant(ant, x)
            if fitness < best_fitness:
                best_fitness = fitness
                best_x = ant.get_x()
                best_ant = ant
        return best_fitness, best_x, best_ant

    def local_pheromone_update(self, ant):
        x_index_and_value = ant.x_index_and_value
        members = ant.members
        for i in range(len(members)):
            j = x_index_and_value[i][0]
            self.pheromone_matrix[i][j] *= self.xi

    def global_pheromone_update_for_one_edge(self, fitness, rank, i, j):
        '''
        :param rank: if rank == 0, this is ant with best fitness value. Max value for rank
         is self.number_of_top_ants
        '''
        self.pheromone_matrix[i][j] += self.rho * (self.number_of_top_ants - rank) / fitness

    def global_pheromone_update(self, ants):
        '''
        Choose best ants based on fitness value and update their pheromone value.
        Number of best ants are self.number_of_top_ants
        '''
        # Sort ants based on their fitness
        sorted_ants = sorted(ants, key=lambda ant: ant.get_fitness())

        for it in range(self.number_of_top_ants):
            x_index_and_value = sorted_ants[it].x_index_and_value
            members = sorted_ants[it].members
            for i in range(len(members)):
                j = x_index_and_value[i][0]
                self.global_pheromone_update_for_one_edge(sorted_ants[it].get_fitness(),
                                                          it, i, j)

    def solve(self, problem, maxiter=100, nmax=10):
        self.penalty = problem.exact_penalty_fun()
        self.problem = problem
        self.initialize()

        # Early stop counter
        N: int = 0
        for t in range(maxiter):
            ants = []

            for k in range(self.number_of_ants):
                ants.append(Ant(self.problem.structure.members))
                ants[k].choose_design_variables(self.probability_matrix, self.design_variables)
                self.local_pheromone_update(ants[k])

            best_fitness, best_x, best_ant = self.evaluate(ants)

            self.global_pheromone_update(ants)
            self.update_probabilities()

            if best_fitness < self.best_fitness:
                self.best_fitness = best_fitness
                self.best_x = best_x
                N = 0
            else:
                N += 1

            if N > nmax:
                return self.best_fitness, self.best_x

        return self.best_fitness, self.best_x

class Ant:
    '''
    Class for storing a single ant
    '''

    def __init__(self, members):

        # list of pairs: pair[0]=index, pair[1] = x = design variables
        self.x_index_and_value = []
        self.members = members

        self.fval = None
        self.penalty = None

    def __choose_design_variable(self, i, probability_matrix, design_variables):
        '''
        Choose design variable (cross section area for truss) based on probability P
        '''
        numerator = probability_matrix[i]
        denominator = np.sum(numerator)
        probabilities = numerator / denominator
        j = np.random.choice(range(len(probabilities)), p=probabilities)
        self.x_index_and_value.append([j, design_variables[j]])

    def choose_design_variables(self, probability_matrix, design_variables):
        '''
        Choose design variable based on probability P for all members.
        '''
        for i in range(len(self.members)):
            self.__choose_design_variable(i, probability_matrix, design_variables)

    def get_x(self):
        '''
        Convert x_index_and_value to np array of x
        :return: design variables x
        '''
        x = map(lambda x: x[1], self.x_index_and_value)
        return np.fromiter(x, dtype=np.float64)

    def get_fitness(self):
        return self.fval + self.penalty

if __name__ == "__main__":
    from metku.optimization.benchmarks import simple_ten_bar_truss

    problem = simple_ten_bar_truss.SimpleTenBarTruss('discrete')

    solver = AntColonyOptimizer(pop_size=100, rho=1, alpha=1,
                                beta=0.5, xi=0.75, number_of_top_ants=4,
                                penalty_factor=10000)

    fopt, xopt = solver.solve(problem=problem, maxiter=100)
    problem(xopt)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Wed Mar 14 19:11:28 2018

EN 1993-1-5 Rules of plated structures

@author: kmela
"""

import math

try:
    from metku.eurocodes.en1993.constants import E, gammaM1
except:
    from eurocodes.en1993.constants import E, gammaM1


def shear_eta(fy):
    """ Computes the 'eta' according to EN 1993-1-5, clause 5.1(2) """

    eta = 1.0

    if fy <= 460.0:
        eta = 1.2

    return eta


def reduction_factor_internal(lp, psi):
    """ Reduction factor for internal compression elements
        4.4(2), and AC:2009
        input: lp .. slenderness (lambda_p)
              psi .. stress ratio
    """

    if lp <= 0.5 + math.sqrt(0.085 - 0.055 * psi):
        rho = 1.0
    else:
        rho = min((lp - 0.055 * (3 + psi)) / lp ** 2, 1.0)

    return rho


def reduction_factor_outstand(lp):
    """ Reduction factor for outstand compression elements
        4.4(2), and AC:2009
        
        input: lp .. slenderness (lambda_p)
    """

    if lp <= 0.748:
        rho = 1.0
    else:
        rho = min((lp - 0.188) / lp ** 2, 1.0)

    return rho


def lambda_p(b, t, epsilon, ksigma):
    """ Slenderness according to 4.4(2)
        input:
            b .. width of the element
            t .. thickness of the element
            epsilon = sqrt(235/fy)
            ksigma .. buckling factor
    """
    return b / t / 28.4 / epsilon / math.sqrt(ksigma)


def buckling_factor_internal(psi=1.0):
    """ Table 4.1 """

    if psi == 1.0:
        ksigma = 4.0
    elif psi > 0:
        ksigma = 8.2 / (1.05 + psi)
    elif psi == 0:
        ksigma = 7.81
    elif psi > -1:
        ksigma = 7.81 - 6.29 * psi + 9.78 * psi ** 2
    elif psi == -1.0:
        ksigma = 23.9
    elif psi > -3:
        ksigma = 5.98 * (1 - psi) ** 2

    return ksigma


def buckling_factor_outstand(psi=1.0, sigma1="tip"):
    """ Table 4.2 
        input:
            sigma1 .. location of the maximum compressive stress
                      "tip" = unsupported end
                      "support" = supported end
    """

    if sigma1 == "tip":
        if psi == 1.0:
            ksigma = 0.43
        elif psi == 0.0:
            ksigma = 0.57
        elif psi == -1:
            ksigma = 0.85
        else:
            ksigma = 0.57 - 0.21 * psi + 0.07 * psi ** 2
    else:
        if psi == 1.0:
            ksigma = 0.43
        if psi < 1.0 and psi > 0.0:
            ksigma = 0.578 / (psi + 0.34)
        elif psi == 0.0:
            ksigma = 1.7
        elif psi < 0 and psi > -1:
            ksigma = 1.7 - 5 * psi + 17.1 * psi ** 2
        else:
            ksigma = 23.8

    return ksigma


def effective_width_internal(b, psi, rho):
    """ EN 1993-1-5 Table 4.1
        Input:
            b .. width of the original plate
            psi .. stress ratio
            rho .. reduction factor for plate buckling
        
        Output:
            beff .. effective width
            be .. be[0] = b_e1, be[1] = b_e2
    """
    be = [0, 0]

    if abs(psi - 1.0) < 1e-6:
        # Psi = 1.0
        beff = rho * b
        be[0] = 0.5 * beff
        be[1] = 0.5 * beff
    elif psi > 0.0:
        beff = rho * b
        be[0] = 2 / (5 - psi) * beff
        be[1] = beff - be[0]
    else:
        # psi < 0
        beff = b * rho / (1 - psi)
        be[0] = 0.4 * beff
        be[1] = 0.6 * beff

    return beff, be


def effective_width_outstand(b, psi, rho):
    """ EN 1993-1-5 Table 4.2
        Input:
            b .. width of the original plate
            psi .. stress ratio
            rho .. reduction factor for plate buckling            
    """

    if psi < 0.0:
        beff = b * rho / (1 - psi)
    else:
        beff = rho * b

    return beff


def transverse_force_resistance(fyw, hw, tw, fyf, bf, tf, ss, a=0, ltype="a",
                                c=0, verb = False):
    """ Clause 6 
        input:
            fyw .. yield strength of the web [MPa]
            hw .. height of the web (mm)
            tw .. web thickness (mm)
            fyf .. yield strength of the flange
            bf .. width of flange
            tf .. thicknes of flange
            a .. distance between transverse stiffeners (a=0 for no stiffeners)
            ss .. length of stiff bearing
            ltype .. type of load application
                    "a" .. through one flange at the middle of the member
                    "b" .. through two flanges at the middle of the member
                    "c" .. through one side of the flange at the end of the member
            c .. distance of the edge of the load application surface from the end
                of the member (for type "c")
    """
    if a == 0:
        a = 1000 * hw

    ss = min(ss, hw)

    if ltype == "a":
        kF = 6 + 2 * (hw / a) ** 2
    elif ltype == "b":
        kF = 3.5 + 2 * (hw / a) ** 2
    elif ltype == "c":
        kF = min(2 + 6 * (ss + c) / hw, 6.0)

    # Next, the reduction factor for effective length (Clause 6.4) and
    # Effective loaded length are determined. This may require iteration
    m1 = fyf * bf / fyw / tw
    m2 = 0

    Fcr = 0.9 * kF * E * tw ** 3 / hw

    if ltype == "a" or ltype == "b":
        ly = min(ss + 2 * tf * (1 + math.sqrt(m1 + m2)), a)
    elif ltype == "c":
        le = min(0.5 * kF * E * tw ** 2 / fyw / hw, ss + c)
        ly = min(le + tf * math.sqrt(0.5 * m1 + (le / tf) ** 2 + m2),
                 le + tf * math.sqrt(m1 + m2))

    lambdaF = math.sqrt(ly * tw * fyw / Fcr)

    if verb:
        print('** Resistance to transverse force **')
        print('   Type (' + ltype + ')')
        print(f'   Length of stiff bearing: ss = {ss:4.2f} mm')
        print(f'   Buckling factor: kF = {kF:4.2f}')
        print(f'   Critical force: Fcr = {Fcr*1e-3:4.2f} kN')
        if ltype == 'c':
            print(f'   Effective load width le = {le:4.2f} mm')
        
        print(f'   m1 = {m1:4.2f}')
        print(f'   m2 = {m2:4.2f}')
        print(f'   ly = {ly:4.2f}')
        print(f'   lambdaF = {lambdaF:4.3f}')
        

    if lambdaF > 0.5:
        # in this case, the initial assumption of m2=0 was wrong, and
        # ly needs to be corrected:
        m2 = 0.02 * (hw / tf) ** 2
        
        if ltype == "a" or ltype == "b":
            ly = min(ss + 2 * tf * (1 + math.sqrt(m1 + m2)), a)
        elif ltype == "c":
            le = min(0.5 * kF * E * tw ** 2 / fyw / hw, ss + c)
            ly = min(le + tf * math.sqrt(0.5 * m1 + (le / tf) ** 2 + m2),
                     le + tf * math.sqrt(m1 + m2))
    
        lambdaF = math.sqrt(ly * tw * fyw / Fcr)
        
        if verb:
            print('Corrected m2 and ly')
            print(f'   m2 = {m2:4.2f}')
            print(f'   ly = {ly:4.2f} mm')
            print(f'   lambdaF = {lambdaF:4.3f}')

    # Reduction factor (Eq. (6.3))
    chiF = min(0.5 / lambdaF, 1.0)

    Leff = chiF * ly

    FRd = fyw * Leff * tw / gammaM1

    if verb:
        print(f' Reduction factor chiF = {chiF:4.2f}')
        print(f' Effective length Leff = {Leff:4.2f} mm')
        print(f' Resistance FRd = {FRd*1e-3:4.2f} kN')

    return FRd


def shear_buckling_reduction_factor(slend_w, eta, end_post="non-rigid",
                                    verb=False):
    """ EN 1993-1-5, Table 5.1 """

    if slend_w < 0.83 / eta:
        chiW = eta
    elif slend_w < 1.08:
        chiW = 0.83 / slend_w
    else:
        if end_post == "non-rigid":
            chiW = 0.83 / slend_w
        elif end_post == "rigid":
            chiW = 1.37 / (0.7 + slend_w)
        else:
            raise ValueError(
                "en1993_1_5.shear_buckling_reduction_factor: bad end_post argument value.")

    if verb:
        print("Shear buckling reduction factor (EN 1993-1-5, Table 5.1)")
        print("End post:" + end_post)
        print("Slenderness: lambda_w = {0:4.3f}".format(slend_w))
        print("Eta: eta = {0:4.3f}".format(eta))
        print("Reduction factor: chi_w = {0:4.3f}".format(chiW))

    return chiW


def shear_buckling_coefficient(hw, a, verb=False):
    """ EN 1993-1-5, A.3
        input: 
            hw .. height of web
            a .. distance between transverse stiffeners
            
        It is assumed that there are no longitudinal stiffeners.
        This has to be modified, if longitudinal stiffeners are included!
    """
    rw = a / hw

    if rw >= 1.0:
        ktau = 5.34 + 4.0 / (rw ** 2)
    else:
        ktau = 4.0 + 5.34 / (rw ** 2)

    if verb:
        print("* Shear buckling coefficient (EN 1993-1-5, A.3): *")
        print("Distance between stiffeners a = {0:4.3f} mm".format(a))
        print("Web height hw = {0:4.3f} mm".format(hw))
        print("Buckling coefficient k_tau = {0:4.3f}".format(ktau))

    return ktau


def sigma_E(t, b):
    """ Critical stress of plate
        EN 1993-1-5, A.1
    """
    return 190_000 * (t / b) ** 2


def tau_crit(hw, a, t, b, verb):
    """ Critical shear stress for shear buckling
        EN 1993-1-5, Eq. (5.4)
    """

    if verb:
        print(
            "* Critical shear stress for shear buckling (EN 1993-1-5, Eq. (5.4)) *")
        print("Web height hw = {0:4.3f}".format(hw))
        print("Distance between stiffeners a = {0:4.3f}".format(a))

    ktau = shear_buckling_coefficient(hw, a, verb)
    sE = sigma_E(t, b)
    tau = ktau * sE

    if verb:
        print("Buckling stress sigma_E = {0:4.3f} MPa".format(sE))
        print("Shear buckling stress tau_crit = {0:4.3f} MPa".format(tau))

    return tau


def shear_buckling_slenderness(fyw, tau_cr, verb=False):
    """ Slenderness for shear buckling
        EN 1993-1-5, Eq. (5.3)
    """

    lambda_w = 0.76 * math.sqrt(fyw / tau_cr)

    if verb:
        print("* Shear buckling slenderness (EN 1993-1-5, Eq. (5.3)) *")
        print("Yield strength of the web: fyw = {0:4.3g}".format(fyw))
        print("Shear buckling stress: tau_crit = {0:4.3f}".format(tau_cr))
        print("Slenderness: lambda_w = {0:4.3f}".format(lambda_w))

    return lambda_w


def shear_buckling_web(chi_w, fyw, hw, t, verb=False):
    """ Shear buckling: contribution from the web
        EN 1993-1-5, Eq. (5.2)
    """

    VbwRd = chi_w * fyw * hw * t / (math.sqrt(3) * gammaM1)

    if verb:
        print("* Part of web: EN 1993-1-5, Eq. (5.2) *")
        print("Web height hw = {0:4.3f} mm".format(hw))
        print("Web thickness tw = {0:4.3f} mm".format(t))
        print("Web yield strength fyw = {0:4.3g} MPa".format(fyw))
        print("Resistance Vbw,Rd = {0:4.3f} kN".format(VbwRd * 1e-3))

    return VbwRd


def shear_buckling_flanges(bf, tf, fyf, a, hw, t, fyw, rM, verb=False):
    """ Shear buckling: contribution from the flanges
        EN 1993-1-5, Eq. (5.2)
    """

    c = a * (0.25 + 1.6 * bf * tf ** 2 * fyf / t / (hw ** 2) / fyw)

    VbfRd = bf * tf ** 2 * fyf / c / gammaM1 * (1 - rM ** 2)

    if verb:
        print("* Part of flanges: EN 1993-1-5, Eq. (5.2) *")
        print("Flange width bf = {0:4.3f} mm".format(bf))
        print("Flange thickness tf = {0:4.3f} mm".format(tf))
        print("Flange yield strength fyw = {0:4.3g} MPa".format(fyf))
        print("c = {0:4.3f} ".format(c))
        print("Resistance Vbf,Rd = {0:4.3f} kN".format(VbfRd * 1e-3))

    return VbfRd



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 21 21:54:04 2021

@author: kmela
"""

from metku.materials.timber_data import Timber, T

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
import matplotlib.path as mpath

class CLTLayer:
    """ Class for CLT slab layers """
    
    def __init__(self,thickness=20,material=Timber(T.C24),orientation=0):
        """
        

        Parameters
        ----------
        thickness : float, optional
            Thickness of the layer. The default is 20.
        material : Timber, optional
            Material propertyies (see timber_data.py). The default is C24.
        orientation : float, optional
            Orientation of the grains. 0 means grains are long the main axis
            and 90 that they are perpendicular to the main axis. The default is 0.

        Returns
        -------
        None.

        """
        
        self.t = thickness
        self.mat = material
        self.orientation = orientation

        self.bx = 1000
        self.by = 1000
    
    def __repr__(self):
        
        return f"({self.t}/{self.mat}/{self.orientation})"
    
    def Ay(self):
        """ Area of the layer with the face in the y-z coordinates """
        
        return self.bx*self.t
    
    def Az(self):
        """ Area of the layer with the face in the x-z coordinates """
        
        return self.by*self.t
    
    def Iy(self):
        """ Second moment of area with respect of the centroid of
            the layer. x-axis is along the span of the slab, and y-axis
            is the horizontal axis
        """
        
        return self.bx*self.t**3/12

    def Iz(self):
        """ Second moment of area with respect of the centroid of
            the layer. x-axis is along the span of the slab, and y-axis
            is the horizontal axis
        """
        
        return self.by*self.t**3/12
    
    def EIy(self):
        """ Bending stiffness of the layer with respect to its
            neutral axis.
        """
        
        if self.orientation == 0:
            E = self.mat.E0mean
        else:
            E = self.mat.E90mean
        
        return E*self.Iy()

    def EIz(self):
        """ Bending stiffness of the layer with respect to its
            neutral axis.
        """
        
        if self.orientation == 0:
            E = self.mat.E90mean
        else:
            E = self.mat.E0mean
        
        return E*self.Iz()


class CLTSlab:
    """ Class for CLT slabs """
    
    def __init__(self,layers,width=1000,length=5000):
        """
        

        Parameters
        ----------
        layers : list of CLTLayer type objects or dict with lists for
                 the following keys: "thickness", "material", "orienatation"
            Layers of the slab.
        width : float
            width of the slab
        length : float
            length of the slab

        Returns
        -------
        None.

        """
        
        self.b = width
        self.L = length
        
        if isinstance(layers,list):
            self.layers = layers
        elif isinstance(layers,dict):
            thickness = layers['thickness']
            materials = layers['material']
            orientations = layers['orientation']
            
            self.layers = []
            
            for t, mat, orient in zip(thickness,materials,orientations):
                self.layers.append(CLTLayer(t,mat,orient))
        
        # Set layer width and length
        for layer in self.layers:
            layer.bx = self.b
            layer.by = self.L
    
    def __repr__(self):
        
        s = 'CLT Slab: '
        
        for i, layer in enumerate(self.layers):
            s += f'{i}: ' + layer.__repr__()
            
            if i < len(self.layers)-1:
                s += ' / '
        
        return s
    
    @property
    def hclt(self):
        """ Total height of the slab """
        
        return sum([layer.t for layer in self.layers])
    
    @property
    def zgc(self):
        """ Location of the centroid in thickness direction """
        return 0.5*self.hclt
    
    @property
    def Ay(self):
        """ Cross-sectional area in y-z-coordinates """
        
        return self.width*self.hclt
    
    @property
    def Az(self):
        """ Cross-sectional area in x-z-coordinates """
        
        return self.length*self.hclt
    
    def EIy(self):
        """ Bending stiffness of the CLT slab with respect to y-axis """
        
        EIy0 = sum([layer.EIy() for layer in self.layers])
        

    def draw(self,axes=None,origin=[0,0],clt_width=0):
        
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        if clt_width == 0:
            clt_width = self.by
        
        h = self.hclt
        for layer in self.layers:
            if layer.orientation == 0:
                layer_hatch = ''
            else:
                layer_hatch = 'x'
            new_patch = patches.Rectangle((origin[0], origin[1]+h-layer.t),
                                       width=clt_width, height=layer.t,
                                       fill=False, hatch=layer_hatch)            
            ax.add_patch(new_patch)
            h -= layer.t
        
        
        ax.set_xlim(origin[0], 100)
        ax.set_ylim(0, self.hclt)
        
        
        ax.set_aspect('equal')
        
        plt.show()
        return ax

if __name__ == '__main__':
    
    layers = {'thickness':[20,20,20],
              'material':[Timber(T.C24),Timber(T.C24),Timber(T.C24)],
              'orientation':[0,90,0]}
    
    s = CLTSlab(layers)

    s.draw()